# License: BSD 3 clause
"""
Functions for running and interacting with SKLL experiments.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Chee Wee Leong (cleong@ets.org)
"""

from __future__ import annotations

import datetime
import json
import logging
from itertools import combinations
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn import __version__ as SCIKIT_VERSION
from sklearn.metrics import get_scorer_names

from skll.config import parse_config_file
from skll.config.utils import _munge_featureset_name
from skll.learner import MAX_CONCURRENT_PROCESSES, Learner, load_custom_learner
from skll.learner.voting import VotingLearner
from skll.metrics import _CUSTOM_METRICS, register_custom_metric
from skll.types import (
    CrossValidateTaskResults,
    EvaluateTaskResults,
    PathOrStr,
    VotingCrossValidateTaskResults,
)
from skll.utils.logging import close_and_remove_logger_handlers, get_skll_logger
from skll.version import __version__

from ..utils.wandb import WandbLogger
from .input import load_featureset
from .output import (
    _print_fancy_output,
    _write_learning_curve_file,
    _write_skll_folds,
    _write_summary_file,
    generate_learning_curve_plots,
)
from .utils import NumpyTypeEncoder, _check_job_results, _create_learner_result_dicts

# Check if gridmap is available
try:
    from gridmap import Job, process_jobs
except ImportError:
    _HAVE_GRIDMAP = False
else:
    _HAVE_GRIDMAP = True

# Turn off interactive plotting for matplotlib
plt.ioff()


__all__ = ["generate_learning_curve_plots", "load_featureset", "run_configuration"]


def _classify_featureset(args: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Run classification job.

    Parameters
    ----------
    args : Dict[str, Any]
        A dictionary with arguments for classifying the ``FeatureSet`` instance.

    Returns
    -------
    res : List[Dict[str, Any]]
        The results of the classification, in the format of a list of dictionaries.

    Raises
    ------
    ValueError
        If extra unknown arguments are passed to the function.
    """
    # Extract all the arguments.
    # (There doesn't seem to be a better way to do this since one can't specify
    # required keyword arguments.)

    experiment_name = args.pop("experiment_name")
    task = args.pop("task")
    sampler = args.pop("sampler")
    feature_hasher = args.pop("feature_hasher")
    hasher_features = args.pop("hasher_features")
    job_name = args.pop("job_name")
    featureset = args.pop("featureset")
    featureset_name = args.pop("featureset_name")
    learner_name = args.pop("learner_name")
    train_path = args.pop("train_path")
    test_path = args.pop("test_path")
    train_set_name = args.pop("train_set_name")
    test_set_name = args.pop("test_set_name")
    shuffle = args.pop("shuffle")
    model_path = args.pop("model_path")
    prediction_prefix = args.pop("prediction_prefix")
    grid_search = args.pop("grid_search")
    grid_objective = args.pop("grid_objective")
    output_metrics = args.pop("output_metrics")
    suffix = args.pop("suffix")
    job_log_file = args.pop("log_file")
    job_log_level = args.pop("log_level")
    probability = args.pop("probability")
    pipeline = args.pop("pipeline")
    results_path = args.pop("results_path")
    fixed_parameters = args.pop("fixed_parameters")
    sampler_parameters = args.pop("sampler_parameters")
    param_grid = args.pop("param_grid")
    pos_label = args.pop("pos_label")
    overwrite = args.pop("overwrite")
    feature_scaling = args.pop("feature_scaling")
    min_feature_count = args.pop("min_feature_count")
    folds_file = args.pop("folds_file")
    grid_search_jobs = args.pop("grid_search_jobs")
    grid_search_folds = args.pop("grid_search_folds")
    cv_folds = args.pop("cv_folds")
    cv_seed = args.pop("cv_seed")
    save_cv_folds = args.pop("save_cv_folds")
    save_cv_models = args.pop("save_cv_models")
    use_folds_file_for_grid_search = args.pop("use_folds_file_for_grid_search")
    stratified_folds = args.pop("do_stratified_folds")
    label_col = args.pop("label_col")
    id_col = args.pop("id_col")
    ids_to_floats = args.pop("ids_to_floats")
    class_map = args.pop("class_map")
    custom_learner_path = args.pop("custom_learner_path")
    custom_metric_path = args.pop("custom_metric_path")
    quiet = args.pop("quiet", False)
    learning_curve_cv_folds = args.pop("learning_curve_cv_folds")
    learning_curve_train_sizes = args.pop("learning_curve_train_sizes")
    save_votes = args.pop("save_votes")

    if args:
        raise ValueError("Extra arguments passed to _classify_featureset: " f"{args.keys()}")
    start_timestamp = datetime.datetime.now()

    # create a new SKLL logger for this specific job and
    # use the given log level
    logger = get_skll_logger(job_name, job_log_file, log_level=job_log_level)

    try:
        # log messages
        logger.info(f"Task: {task}")

        # check if we have any possible custom metrics
        possible_custom_metric_names = []
        for metric_name in output_metrics + [grid_objective]:
            # metrics that are not in `SCORERS` or `None` are candidates
            # (the `None` is a by-product of how jobs with single tuning
            # objectives are created)
            if (
                metric_name not in get_scorer_names()
                and metric_name not in _CUSTOM_METRICS
                and metric_name is not None
            ):
                possible_custom_metric_names.append(metric_name)

        # initialize list that will hold any invalid metrics
        # that we could not register as custom metrics
        invalid_metric_names = []

        # if we have possible custom metrics
        if possible_custom_metric_names:
            # check that we have a file to load them from
            if not custom_metric_path:
                raise ValueError(f"invalid metrics specified: {possible_custom_metric_names}")
            else:
                # try to register each possible custom metric
                # raise an exception if we fail, if we don't then
                # add the custom metric function to `globals()` so
                # that it serializes properly for gridmap
                for custom_metric_name in possible_custom_metric_names:
                    try:
                        custom_metric_func = register_custom_metric(
                            custom_metric_path, custom_metric_name
                        )
                    except (AttributeError, NameError, ValueError):
                        invalid_metric_names.append(custom_metric_name)
                    else:
                        logger.info(f"registered '{custom_metric_name}' as a " f"custom metric")
                        globals()[custom_metric_name] = custom_metric_func

        # raise an error if we have any invalid metrics
        if invalid_metric_names:
            raise ValueError(
                f"invalid metrics specified: {invalid_metric_names}. "
                f"If these are custom metrics, check the function "
                f"names."
            )

        if task == "cross_validate":
            if isinstance(cv_folds, int):
                num_folds = cv_folds
            else:  # folds_file was used, so count the unique fold ids.
                num_folds = len(set(cv_folds.values()))
            logger.info(
                f"Cross-validating ({num_folds} folds, seed={cv_seed}) on "
                f"{train_set_name}, feature set {featureset} ..."
            )
        elif task == "evaluate":
            logger.info(
                f"Training on {train_set_name}, Test on "
                f"{test_set_name}, feature set {featureset} ..."
            )
        elif task == "train":
            logger.info(f"Training on {train_set_name}, feature set " f"{featureset} ...")
        elif task == "learning_curve":
            logger.info(
                "Generating learning curve "
                f"({learning_curve_cv_folds} 80/20 folds, "
                f"sizes={learning_curve_train_sizes}, "
                f"objective={grid_objective}) on {train_set_name}, "
                f"feature set {featureset} ..."
            )
        else:  # predict
            logger.info(
                f"Training on {train_set_name}, Making predictions on"
                f" {test_set_name}, feature set {featureset} ..."
            )

        # check whether a trained model on the same data with the same
        # featureset already exists if so, load it and then use it on test data
        modelfile = Path(model_path) / f"{job_name}.model"
        train_set_size: Union[int, str] = "unknown"
        if task in ["cross_validate", "learning_curve"] or not modelfile.exists() or overwrite:
            train_examples = load_featureset(
                train_path,
                featureset,
                suffix,
                label_col=label_col,
                id_col=id_col,
                ids_to_floats=ids_to_floats,
                quiet=quiet,
                class_map=class_map,
                feature_hasher=feature_hasher,
                num_features=hasher_features,
                logger=logger,
            )

            train_set_size = len(train_examples.ids)

            if not train_examples.has_labels:
                raise ValueError("Training examples do not have labels")

            # set up some keyword arguments for instantiating the learner
            # object; note that these are shared by all types of learners
            # supported by SKLL (regular and voting)
            common_learner_kwargs = {
                "custom_learner_path": custom_learner_path,
                "feature_scaling": feature_scaling,
                "pos_label": pos_label,
                "min_feature_count": min_feature_count,
                "logger": logger,
            }

            # instantiate the right type of learner object
            learner: Union[Learner, VotingLearner]
            if learner_name in ["VotingClassifier", "VotingRegressor"]:
                # the fixed parameters dictionary must at least
                # contains the estimator names for the voting;
                # the rest can be set to default values
                try:
                    learner_names = fixed_parameters["estimator_names"]
                except KeyError:
                    raise ValueError(
                        "'estimator_names' must be specified as "
                        "fixed parameters for voting classifiers "
                        "and/or regressors."
                    ) from None
                else:
                    voting_type = fixed_parameters.get("voting_type", "hard")
                    model_kwargs_list = fixed_parameters.get("estimator_fixed_parameters")
                    sampler_list = fixed_parameters.get("estimator_samplers")
                    sampler_kwargs_list = fixed_parameters.get("estimator_sampler_parameters")
                    param_grids_list = fixed_parameters.get("estimator_param_grids")
                    learner = VotingLearner(
                        learner_names,
                        voting=voting_type,
                        model_kwargs_list=model_kwargs_list,
                        sampler_list=sampler_list,
                        sampler_kwargs_list=sampler_kwargs_list,
                        **common_learner_kwargs,
                    )
            else:
                learner = Learner(
                    learner_name,
                    model_kwargs=fixed_parameters,
                    pipeline=pipeline,
                    sampler=sampler,
                    sampler_kwargs=sampler_parameters,
                    probability=probability,
                    **common_learner_kwargs,
                )

        # load the model if it already exists
        else:
            # import custom learner into global namespace if we are reusing
            # a saved model
            if custom_learner_path:
                globals()[learner_name] = load_custom_learner(custom_learner_path, learner_name)

            # load the non-custom learner from disk
            if modelfile.exists() and not overwrite:
                logger.info(f"Loading pre-existing {learner_name} model: " f"{modelfile}")
            if learner_name in ["VotingClassifier", "VotingRegressor"]:
                learner = VotingLearner.from_file(modelfile, logger=logger)
            else:
                learner = Learner.from_file(modelfile, logger=logger)

        # Load test set if there is one
        test_set_size: Union[int, str] = "n/a"
        if task == "evaluate" or task == "predict":
            test_examples = load_featureset(
                test_path,
                featureset,
                suffix,
                label_col=label_col,
                id_col=id_col,
                ids_to_floats=ids_to_floats,
                quiet=quiet,
                class_map=class_map,
                feature_hasher=feature_hasher,
                num_features=hasher_features,
            )
            test_set_size = len(test_examples.ids)

        # compute information about xval and grid folds that can be put in results
        # in readable form
        if isinstance(cv_folds, dict):
            cv_folds_to_print = f"{len(set(cv_folds.values()))} via folds file"
        else:
            cv_folds_to_print = str(cv_folds)

        if isinstance(grid_search_folds, dict):
            grid_search_folds_to_print = f"{len(set(grid_search_folds.values()))} via folds file"
        else:
            grid_search_folds_to_print = str(grid_search_folds)

        # create a list of dictionaries of the results information
        learner_result_dict_base = {
            "experiment_name": experiment_name,
            "job_name": job_name,
            "train_set_name": train_set_name,
            "train_set_size": train_set_size,
            "test_set_name": test_set_name,
            "test_set_size": test_set_size,
            "featureset": json.dumps(featureset),
            "featureset_name": featureset_name,
            "shuffle": shuffle,
            "learner_name": learner_name,
            "task": task,
            "start_timestamp": start_timestamp.strftime("%d %b %Y %H:%M:" "%S.%f"),
            "version": __version__,
            "feature_scaling": feature_scaling,
            "folds_file": folds_file,
            "grid_search": grid_search,
            "grid_objective": grid_objective,
            "grid_search_folds": grid_search_folds_to_print,
            "min_feature_count": min_feature_count,
            "cv_folds": cv_folds_to_print,
            "using_folds_file": isinstance(cv_folds, dict) or isinstance(grid_search_folds, dict),
            "save_cv_folds": save_cv_folds,
            "save_cv_models": save_cv_models,
            "use_folds_file_for_grid_search": use_folds_file_for_grid_search,
            "stratified_folds": stratified_folds,
            "scikit_learn_version": SCIKIT_VERSION,
        }

        # check if we're doing cross-validation, because we only load/save
        # models when we're not.
        task_results: List[EvaluateTaskResults]
        grid_scores: Union[List[None], List[float]] = [None]
        grid_search_cv_results_dicts: Union[List[None], List[Dict[str, Any]]] = [None]

        if task == "cross_validate":
            logger.info("Cross-validating")

            # set up the keyword arguments for learner cross-validation;
            # note that most are shared by all types of learners
            # supported by SKLL (regular and voting)
            xval_kwargs = {
                "shuffle": shuffle,
                "stratified": stratified_folds,
                "prediction_prefix": prediction_prefix,
                "grid_search": grid_search,
                "grid_search_folds": grid_search_folds,
                "cv_folds": cv_folds,
                "cv_seed": cv_seed,
                "grid_objective": grid_objective,
                "output_metrics": output_metrics,
                "grid_jobs": grid_search_jobs,
                "save_cv_folds": save_cv_folds,
                "save_cv_models": save_cv_models,
                "use_custom_folds_for_grid_search": use_folds_file_for_grid_search,
            }

            # voting learners require an optional list of parameter grids
            # passed as fixed parameters whereas regular learners only need
            # a single parameter grid
            if isinstance(learner, VotingLearner):
                xval_kwargs["param_grid_list"] = param_grids_list
                xval_kwargs["individual_predictions"] = save_votes
            else:
                xval_kwargs["param_grid"] = param_grid

            # cross-validate the learner
            # voting learners return only a subset of the results
            # for cross-validation (no grid search results)
            models: Union[Optional[List[Learner]], Optional[List[VotingLearner]]]
            if isinstance(learner, Learner):
                learner_cv_results: CrossValidateTaskResults = learner.cross_validate(
                    train_examples, **xval_kwargs
                )
                (
                    task_results,
                    grid_scores,
                    grid_search_cv_results_dicts,
                    skll_fold_ids,
                    models,
                ) = learner_cv_results
            else:
                voting_learner_cv_results: VotingCrossValidateTaskResults = learner.cross_validate(
                    train_examples, **xval_kwargs
                )
                (task_results, skll_fold_ids, models) = voting_learner_cv_results
                grid_scores = [None] * cv_folds
                grid_search_cv_results_dicts = [None] * cv_folds

            if models and len(models) > 0:
                for index, model in enumerate(models, start=1):
                    assert isinstance(model, (Learner, VotingLearner))
                    modelfile = Path(model_path) / f"{job_name}_fold{index}.model"
                    model.save(modelfile)

        elif task == "learning_curve":
            logger.info("Generating learning curve(s)")
            (
                curve_train_scores,
                curve_test_scores,
                curve_fit_times,
                computed_curve_train_sizes,
            ) = learner.learning_curve(
                train_examples,
                grid_objective,
                cv_folds=learning_curve_cv_folds,
                train_sizes=learning_curve_train_sizes,
            )
        else:
            # if we do not have a saved model, we need to train one
            if not modelfile.exists() or overwrite:
                logger.info(f"Featurizing and training new {learner_name} model")

                # set up the keyword arguments for learner training;
                # note that most are shared by all types of learners
                # supported by SKLL (regular and voting)
                train_kwargs = {
                    "grid_search": grid_search,
                    "grid_search_folds": grid_search_folds,
                    "grid_objective": grid_objective,
                    "grid_jobs": grid_search_jobs,
                    "shuffle": shuffle,
                }

                # voting learners require an optional list of parameter grids
                # passed as fixed parameters whereas regular learners only need
                # a single parameter grid
                if isinstance(learner, VotingLearner):
                    train_kwargs["param_grid_list"] = param_grids_list
                else:
                    train_kwargs["param_grid"] = param_grid

                # train the model

                # regular learners return a grid score and results
                if isinstance(learner, Learner):
                    results = learner.train(train_examples, **train_kwargs)
                    grid_scores = [results[0]]
                    grid_search_cv_results_dicts = [results[1]]
                    if grid_search:
                        logger.info(
                            f"Best {grid_objective} grid search score: " f"{round(results[0], 3)}"
                        )
                else:
                    learner.train(train_examples, **train_kwargs)

                # save model, if asked
                if model_path:
                    learner.save(modelfile)
                    learner_result_dict_base["model_file"] = str(modelfile)

            # print out the model parameters; note that for
            # voting learners, we exclude the parameters for
            # the underlying estimators
            params = learner.model.get_params(deep=isinstance(learner, Learner))
            param_out = (
                f"{param_name}: {param_value}" for param_name, param_value in params.items()
            )
            logger.info(f"Hyperparameters: {', '.join(param_out)}")

            # evaluate the model on the test set;
            if task == "evaluate":
                logger.info("Evaluating predictions")
                # voting learners have an extra keyword argument indicating
                # whether we want to save the individual predictions (or votes)
                extra_kwargs = {}
                if isinstance(learner, VotingLearner):
                    extra_kwargs["individual_predictions"] = save_votes
                task_results = [
                    learner.evaluate(
                        test_examples,
                        prediction_prefix=prediction_prefix,
                        grid_objective=grid_objective,
                        output_metrics=output_metrics,
                        **extra_kwargs,
                    )
                ]
            elif task == "predict":
                logger.info("Writing predictions")
                # we set `class_labels` to `False` so that if the learner is
                # probabilistic, probabilities are written instead of labels;
                # voting learners have an extra keyword argument indicating
                # whether we want to save the individual predictions (or votes)
                extra_kwargs = {}
                if isinstance(learner, VotingLearner):
                    extra_kwargs["individual_predictions"] = save_votes
                learner.predict(
                    test_examples,
                    prediction_prefix=prediction_prefix,
                    class_labels=False,
                    **extra_kwargs,
                )
                learner_result_dict_base[
                    "predictions_file"
                ] = f"{prediction_prefix}_predictions.tsv"

        end_timestamp = datetime.datetime.now()
        learner_result_dict_base["end_timestamp"] = end_timestamp.strftime("%d %b %Y %H:%M:%S.%f")
        total_time = end_timestamp - start_timestamp
        learner_result_dict_base["total_time"] = str(total_time)

        results_json_path = Path(results_path) / f"{job_name}.results.json"

        if task == "cross_validate" or task == "evaluate":
            res = _create_learner_result_dicts(
                task_results, grid_scores, grid_search_cv_results_dicts, learner_result_dict_base
            )

            # write out the result dictionary to a json file
            with open(results_json_path, "w") as json_file:
                json.dump(res, json_file, cls=NumpyTypeEncoder)

            with open(Path(results_path) / f"{job_name}.results", "w") as output_file:
                _print_fancy_output(res, output_file)

        elif task == "learning_curve":
            result_dict = {}
            result_dict.update(learner_result_dict_base)
            result_dict.update(
                {
                    "learning_curve_cv_folds": learning_curve_cv_folds,
                    "given_curve_train_sizes": learning_curve_train_sizes,
                    "learning_curve_train_scores_means": np.mean(curve_train_scores, axis=1),
                    "learning_curve_test_scores_means": np.mean(curve_test_scores, axis=1),
                    "learning_curve_fit_times_means": np.mean(curve_fit_times, axis=1),
                    "learning_curve_train_scores_stds": np.std(curve_train_scores, axis=1, ddof=1),
                    "learning_curve_test_scores_stds": np.std(curve_test_scores, axis=1, ddof=1),
                    "learning_curve_fit_times_stds": np.std(curve_fit_times, axis=1),
                    "computed_curve_train_sizes": computed_curve_train_sizes,
                }
            )

            # we need to return and write out a list of dictionaries
            res = [result_dict]

            # write out the result dictionary to a json file
            with open(results_json_path, "w") as json_file:
                json.dump(res, json_file, cls=NumpyTypeEncoder)

        # For all other tasks, i.e. train or predict
        else:
            if results_path:
                assert len(grid_scores) == 1
                assert len(grid_search_cv_results_dicts) == 1
                grid_search_cv_results_dict: Dict[str, Any] = {
                    "grid_score": grid_scores[0],
                    "grid_search_cv_results": grid_search_cv_results_dicts[0],
                }
                grid_search_cv_results_dict.update(learner_result_dict_base)
                # write out the result dictionary to a json file
                with open(results_json_path, "w") as json_file:
                    json.dump(grid_search_cv_results_dict, json_file, cls=NumpyTypeEncoder)
            res = [learner_result_dict_base]

        # write out the cv folds if required
        if task == "cross_validate" and save_cv_folds and skll_fold_ids is not None:
            skll_fold_ids_file = f"{experiment_name}_skll_fold_ids.csv"
            with open(Path(results_path) / skll_fold_ids_file, "w") as output_file:
                _write_skll_folds(skll_fold_ids, output_file)

    finally:
        close_and_remove_logger_handlers(logger)

    return res


def run_configuration(
    config_file: PathOrStr,
    local: bool = False,
    overwrite: bool = True,
    queue: str = "all.q",  # noqa: C901
    hosts: Optional[List[str]] = None,
    write_summary: bool = True,
    quiet: bool = False,
    ablation: int = 0,
    resume: bool = False,
    log_level: int = logging.INFO,
) -> List[str]:
    """
    Run jobs specified in the configuration file locally or on the grid.

    Parameters
    ----------
    config_file : :class:`skll.types.PathOrStr`
        Path to the configuration file we would like to use.
    local : bool, default=False
        Should this be run locally instead of on the cluster?
    overwrite : bool, default=True
        If the model files already exist, should we overwrite
        them instead of re-using them?
    queue : str, default="all.q"
        The DRMAA queue to use if we're running on the cluster.
    hosts : Optional[List[str]], default=None
        If running on the cluster, these are the machines we should use.
    write_summary : bool, default=True
        Write a TSV file with a summary of the results.
    quiet : bool, default=False
        Suppress printing of "Loading..." messages.
    ablation : int, default=0
        Number of features to remove when doing an ablation
        experiment. If positive, we will perform repeated ablation
        runs for all combinations of features removing the
        specified number at a time. If ``None``, we will use all
        combinations of all lengths. If 0, the default, no
        ablation is performed. If negative, a ``ValueError`` is
        raised.
    resume : bool, default=False
        If result files already exist for an experiment, do not
        overwrite them. This is very useful when doing a large
        ablation experiment and part of it crashes.
    log_level : int, default=logging.INFO
        The level for logging messages.

    Returns
    -------
    result_json_paths : List[str]
        A list of paths to .json results files for each variation in the
        experiment.

    Raises
    ------
    ValueError
        If value for ``"ablation"`` is not a positive int or ``None``.
    OSError
        If the lenth of the ``FeatureSet`` name > 210.
    """
    try:
        # Read configuration
        (
            experiment_name,
            task,
            sampler,
            fixed_sampler_parameters,
            feature_hasher,
            hasher_features,
            id_col,
            label_col,
            train_set_name,
            test_set_name,
            suffix,
            featuresets,
            do_shuffle,
            model_path,
            do_grid_search,
            grid_objectives,
            probability,
            pipeline,
            results_path,
            pos_label,
            feature_scaling,
            min_feature_count,
            folds_file,
            grid_search_jobs,
            grid_search_folds,
            cv_folds,
            cv_seed,
            save_cv_folds,
            save_cv_models,
            use_folds_file_for_grid_search,
            do_stratified_folds,
            fixed_parameter_list,
            param_grid_list,
            featureset_names,
            learners,
            prediction_dir,
            log_path,
            train_path,
            test_path,
            ids_to_floats,
            class_map,
            custom_learner_path,
            custom_metric_path,
            learning_curve_cv_folds_list,
            learning_curve_train_sizes,
            output_metrics,
            save_votes,
            wandb_credentials,
        ) = parse_config_file(config_file, log_level=log_level)

        # get the main experiment logger that will already have been
        # created by the configuration parser so we don't need anything
        # except the name `experiment`.
        logger = get_skll_logger("experiment")
        # init WandbLogger. If W&B credentials are not specified,
        # logging to wandb will not be performed.
        wandb_logger = WandbLogger(wandb_credentials, str(config_file))

        # Check if we have gridmap
        if not local and not _HAVE_GRIDMAP:
            local = True
            logger.warning(
                "gridmap 0.10.1+ not available. Forcing local "
                "mode.  To run things on a DRMAA-compatible "
                "cluster, install gridmap>=0.10.1 via pip."
            )

        # No grid search or ablation for learning curve generation
        if task == "learning_curve":
            if ablation is None or ablation > 0:
                ablation = 0
                logger.warning(
                    "Ablating features is not supported during "
                    "learning curve generation. Ignoring."
                )

        # if we just had a train file and a test file, there are no real featuresets
        # in which case there are no features to ablate
        if len(featuresets) == 1 and len(featuresets[0]) == 1:
            if ablation is None or ablation > 0:
                ablation = 0
                logger.warning("Not enough featuresets for ablation. Ignoring.")

        # if performing ablation, expand featuresets to include combinations of
        # features within those sets
        if ablation is None or ablation > 0:
            # Make new feature set lists so that we can iterate without issue
            expanded_fs = []
            expanded_fs_names = []
            for featureset_prefix_list, featureset_name in zip(featuresets, featureset_names):
                featureset_prefix_sorted_list = sorted(featureset_prefix_list)
                featureset_prefix_set = set(featureset_prefix_sorted_list)
                # Expand to all feature combinations if ablation is None
                if ablation is None:
                    for i in range(1, len(featureset_prefix_sorted_list)):
                        for excluded_features in combinations(featureset_prefix_sorted_list, i):
                            expanded_fs.append(
                                sorted(featureset_prefix_set - set(excluded_features))
                            )
                            expanded_fs_names.append(
                                f"{featureset_name}_minus_"
                                f"{_munge_featureset_name(excluded_features)}"
                            )
                # Otherwise, just expand removing the specified number at a time
                else:
                    for excluded_features in combinations(featureset_prefix_sorted_list, ablation):
                        expanded_fs.append(sorted(featureset_prefix_set - set(excluded_features)))
                        expanded_fs_names.append(
                            f"{featureset_name}_minus_"
                            f"{_munge_featureset_name(excluded_features)}"
                        )
                # Also add version with nothing removed as baseline
                expanded_fs.append(featureset_prefix_sorted_list)
                expanded_fs_names.append(f"{featureset_name}_all")

            # Replace original feature set lists
            featuresets = expanded_fs
            featureset_names = expanded_fs_names
        elif ablation < 0:
            raise ValueError(
                'Value for "ablation" argument must be either ' "positive integer or None."
            )

        # the list of jobs submitted (if running on grid)
        if not local:
            jobs = []

        # the list to hold the paths to all the result json files
        result_json_paths = []

        # check if the length of the featureset_name exceeds the maximum length
        # allowed
        for featureset_name in featureset_names:
            if len(featureset_name) > 210:
                raise OSError(
                    f'System generated file length "{featureset_name}" '
                    "exceeds the maximum length supported.  Please specify "
                    'names of your datasets with "featureset_names".  If you '
                    "are running ablation experiment, please reduce the "
                    'length of the features in "featuresets" because the '
                    "auto-generated name would be longer than the file system"
                    " can handle"
                )

        # if the task is learning curve, and ``metrics`` was specified, then
        # assign the value of ``metrics`` to ``grid_objectives`` - this lets
        # us piggyback on the parallelization of the objectives that is already
        # set up for us to use
        if task == "learning_curve" and len(output_metrics) > 0:
            grid_objectives = output_metrics

        # if there were no grid objectives provided, just set it to
        # a list containing a single None to allow the parallelization
        # to proceed and to pass the correct default value of grid_objective
        # down to _classify_featureset().
        grid_objectives_extra: Union[List[str], List[None]]
        grid_objectives_extra = [None] if not grid_objectives else grid_objectives

        job_results = []
        # Run each featureset-learner-objective combination
        for featureset, featureset_name in zip(featuresets, featureset_names):
            for learner_num, learner_name in enumerate(learners):
                for grid_objective in grid_objectives_extra:
                    # The individual job name is built from the experiment name, featureset,
                    # learner and objective. To keep the job name as compact as possible,
                    # some of the components are only included if they have multiple values
                    # in this run.
                    job_name_components = [experiment_name]
                    if len(featuresets) > 1:
                        job_name_components.append(featureset_name)

                    job_name_components.append(learner_name)

                    if len(grid_objectives_extra) > 1 and grid_objective is not None:
                        job_name_components.append(grid_objective)

                    job_name = "_".join(job_name_components)

                    logger.info(f"Job name: {job_name}")
                    wandb_logger.log_to_summary(job_name, "featureset name", featureset_name)
                    wandb_logger.log_to_summary(job_name, "learner name", learner_name)
                    wandb_logger.log_to_summary(job_name, "grid objective", grid_objective)

                    # change the prediction prefix to include the feature set
                    prediction_prefix = str(Path(prediction_dir) / job_name)

                    # the log file that stores the actual output of this script (e.g.,
                    # the tuned parameters, what kind of experiment was run, etc.)
                    logfile = Path(log_path) / f"{job_name}.log"

                    # Figure out result json file path
                    result_json_path = Path(results_path) / f"{job_name}.results.json"

                    # save the path to the results json file that will be written
                    result_json_paths.append(str(result_json_path))

                    # If result file already exists and we're resuming, move on
                    if resume and (result_json_path.exists() and result_json_path.stat().st_size):
                        logger.info(
                            "Running in resume mode and "
                            f"{result_json_path} exists, so skipping "
                            "job."
                        )
                        continue

                    # create args for classification job
                    job_args: Dict[str, Any] = {}
                    job_args["experiment_name"] = experiment_name
                    job_args["task"] = task
                    job_args["sampler"] = sampler
                    job_args["feature_hasher"] = feature_hasher
                    job_args["hasher_features"] = hasher_features
                    job_args["job_name"] = job_name
                    job_args["featureset"] = featureset
                    job_args["featureset_name"] = featureset_name
                    job_args["learner_name"] = learner_name
                    job_args["train_path"] = train_path
                    job_args["test_path"] = test_path
                    job_args["train_set_name"] = train_set_name
                    job_args["test_set_name"] = test_set_name
                    job_args["shuffle"] = do_shuffle
                    job_args["model_path"] = model_path
                    job_args["prediction_prefix"] = prediction_prefix
                    job_args["grid_search"] = do_grid_search
                    job_args["grid_objective"] = grid_objective
                    job_args["output_metrics"] = output_metrics
                    job_args["suffix"] = suffix
                    job_args["log_file"] = logfile
                    job_args["log_level"] = log_level
                    job_args["probability"] = probability
                    job_args["pipeline"] = pipeline
                    job_args["save_votes"] = save_votes
                    job_args["results_path"] = results_path
                    job_args["sampler_parameters"] = (
                        fixed_sampler_parameters if fixed_sampler_parameters else dict()
                    )
                    job_args["fixed_parameters"] = (
                        fixed_parameter_list[learner_num] if fixed_parameter_list else dict()
                    )
                    job_args["param_grid"] = (
                        param_grid_list[learner_num] if param_grid_list else None
                    )
                    job_args["pos_label"] = pos_label
                    job_args["overwrite"] = overwrite
                    job_args["feature_scaling"] = feature_scaling
                    job_args["min_feature_count"] = min_feature_count
                    job_args["grid_search_jobs"] = grid_search_jobs
                    job_args["grid_search_folds"] = grid_search_folds
                    job_args["folds_file"] = folds_file
                    job_args["cv_folds"] = cv_folds
                    job_args["cv_seed"] = cv_seed
                    job_args["save_cv_folds"] = save_cv_folds
                    job_args["save_cv_models"] = save_cv_models
                    job_args["use_folds_file_for_grid_search"] = use_folds_file_for_grid_search
                    job_args["do_stratified_folds"] = do_stratified_folds
                    job_args["label_col"] = label_col
                    job_args["id_col"] = id_col
                    job_args["ids_to_floats"] = ids_to_floats
                    job_args["quiet"] = quiet
                    job_args["class_map"] = class_map
                    job_args["custom_learner_path"] = custom_learner_path
                    job_args["custom_metric_path"] = custom_metric_path
                    job_args["learning_curve_cv_folds"] = learning_curve_cv_folds_list[learner_num]
                    job_args["learning_curve_train_sizes"] = learning_curve_train_sizes

                    if not local:
                        # add to job list if we're doing things on the grid
                        jobs.append(
                            Job(
                                _classify_featureset,
                                [job_args],
                                num_slots=(
                                    MAX_CONCURRENT_PROCESSES
                                    if (do_grid_search or task == "learning_curve")
                                    else 1
                                ),
                                name=job_name,
                                queue=queue,
                            )
                        )
                    else:
                        job_results.append(_classify_featureset(job_args))

        # Call get_skll_logger again after _classify_featureset
        # calls are finished so that any warnings that may
        # happen after this point get correctly logged to the
        # main logger
        logger = get_skll_logger("experiment")

        # submit the jobs (if running on grid)
        if not local and _HAVE_GRIDMAP:
            if log_path:
                job_results = process_jobs(jobs, white_list=hosts, temp_dir=log_path)
            else:
                job_results = process_jobs(jobs, white_list=hosts)
            _check_job_results(job_results)

        # process output and log to wandb
        if task == "predict":
            for job_result in job_results:
                task_result = job_result[0]  # predict outputs a single dict
                wandb_logger.log_predict_results(task_result)

        elif task == "train":
            for job_result in job_results:
                # train outputs a single dict
                wandb_logger.log_train_results(job_result[0])

        elif task == "cross_validate" or task == "evaluate":
            if write_summary:
                # write out the summary results file
                summary_file_path = Path(results_path) / f"{experiment_name}_summary.tsv"
                with open(summary_file_path, "w", newline="") as output_file:
                    _write_summary_file(result_json_paths, output_file, ablation=ablation)
                wandb_logger.log_summary_file(summary_file_path)

            for job_result_list in job_results:
                for result_dict in job_result_list:
                    wandb_logger.log_evaluation_results(result_dict)
            wandb_logger.log_label_metric_table()

        elif task == "learning_curve":
            # write out the summary file
            summary_file_path = Path(results_path) / f"{experiment_name}_summary.tsv"
            with open(summary_file_path, "w", newline="") as output_file:
                _write_learning_curve_file(result_json_paths, output_file)
            wandb_logger.log_summary_file(summary_file_path)

            # generate the actual plot if we have the requirements installed
            plot_paths = generate_learning_curve_plots(
                experiment_name, results_path, summary_file_path
            )
            for plot_path in plot_paths:
                wandb_logger.log_plot(plot_path)

    finally:
        # Close/remove any logger handlers
        close_and_remove_logger_handlers(get_skll_logger("experiment"))

    return result_json_paths



"""
Utility classes and functions for logging to Weights & Biases.

:author: Tamar Lavee (tlavee@ets.org)
"""

from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd

import wandb
from skll.config import _setup_config_parser
from skll.types import PathOrStr
from wandb.sdk.lib import RunDisabled
from wandb.sdk.wandb_run import Run


class WandbLogger:
    """Interface for Weights and Biases logging."""

    def __init__(self, wandb_credentials: Optional[Dict[str, str]], config_file_path: str):
        """
        Initialize the wandb_run if wandb_credentials are provided.

        Parameters
        ----------
        wandb_credentials : Optional[Dict[str, str]]
            A dictionary containing the W&B entity and project names that will be
            used to initialize the wandb run. If ``None``, logging to W&B will not be performed.
        config_file_path : str
            The path to this experiment's config file
        """
        self.wandb_run: Optional[Union[Run, RunDisabled]] = None
        if wandb_credentials:
            self.wandb_run = wandb.init(
                project=wandb_credentials["wandb_project"],
                entity=wandb_credentials["wandb_entity"],
                config=get_config_dict(config_file_path),
            )
        self.label_metrics_table: Optional[wandb.Table] = None

    def log_plot(self, plot_file_path: str) -> None:
        """
        Log a plot to W&B as an image if logging to W&B is enabled.

        Parameters
        ----------
        plot_file_path : str
            The full path to the plot file.
        """
        plot_name = Path(plot_file_path).stem
        if self.wandb_run:
            self.wandb_run.log({plot_name: wandb.Image(plot_file_path)})

    def log_summary_file(self, summary_file_path: PathOrStr) -> None:
        """
        Log a task summary file to W&B if logging to W&B is enabled.

        The summary is logged as a table to the wandb run.

        Parameters
        ----------
        summary_file_path : PathOrStr
            The path to the summary tsv file
        """
        if self.wandb_run:
            summary_df = pd.read_csv(summary_file_path, sep="\t")
            summary_table = wandb.Table(dataframe=summary_df, allow_mixed_types=True)
            self.wandb_run.log({"Summary": summary_table})

    def log_evaluation_results(self, task_results: Dict[str, Any]) -> None:
        """
        Log evaluation results to W&B, if logging to W&B is enabled.

        Log basic task info, general scores, and label-specific evaluation scores
        and confusion matrix for classification tasks.

        The input is either "evaluate" task results or a single
        fold results in a "cross validate" task.

        The metrics are logged to a section named by the job name and
        fold, if available.

        Parameters
        ----------
        task_results : Dict[str,Any]
            The evaluation results of a single job of "evaluate" task or
            a single fold of a "cross_validate" task.
        """
        if self.wandb_run:
            task_prefix = task_results["job_name"]
            # if this is a fold's result, add fold name to the prefix.
            if fold_num := task_results.get("fold"):
                task_prefix = f"{task_prefix}_fold_{fold_num}"

            # log confusion matrix as a custom chart
            if confusion_matrix := task_results.get("conf_matrix"):
                self.log_conf_matrix_chart(
                    task_prefix, confusion_matrix, sorted(task_results["label_metrics"].keys())
                )

                # log Precision, recall and f-measure for each label
                if not self.label_metrics_table:
                    self.label_metrics_table = wandb.Table(
                        columns=["Job Name", "Label", "Precision", "Recall", "F-measure"],
                        allow_mixed_types=True,
                    )
                for label, label_metric_dict in task_results["label_metrics"].items():
                    self.label_metrics_table.add_data(
                        task_prefix,
                        label,
                        label_metric_dict.get("precision"),
                        label_metric_dict.get("recall"),
                        label_metric_dict.get("f-measure"),
                    )

    def log_label_metric_table(self) -> None:
        """
        Log the full label metric table to W&B, if logging to W&B is enabled.

        This table is populated when evaluation results are logged for
        classification learners. This method should be called after all task
        results of a `evaluate` or `cross_validate` task have been logged.
        """
        if self.wandb_run and self.label_metrics_table:
            self.wandb_run.log({"classification metrics": self.label_metrics_table})

    def log_train_results(self, task_results: Dict[str, Any]) -> None:
        """
        Log train results to W&B, if logging to W&B is enabled.

        Log basic task info and the model file location.

        Parameters
        ----------
        task_results : Dict[str, Any]
            The train task results.
        """
        if self.wandb_run:
            task_prefix = task_results["job_name"]
            for metric in ["train_set_size", "test_set_size", "model_file"]:
                self.log_to_summary(task_prefix, metric, task_results.get(metric, "N/A"))

    def log_predict_results(self, task_results: Dict[str, Any]) -> None:
        """
        Log predict results to W&B, if logging to W&B is enabled.

        Log basic task info to the run, and predictions file as a table.

        Parameters
        ----------
        task_results : Dict[str, Any]
            The predict task results.
        """
        if self.wandb_run:
            task_prefix = task_results["job_name"]
            for metric in ["train_set_size", "test_set_size"]:
                self.log_to_summary(task_prefix, metric, task_results.get(metric, "N/A"))
            predictions_df = pd.read_csv(task_results["predictions_file"], sep="\t")
            predictions_table = wandb.Table(dataframe=predictions_df, allow_mixed_types=True)
            self.wandb_run.log({f"{task_prefix}/predictions": predictions_table})

    def log_conf_matrix_chart(self, task_prefix, confusion_matrix, labels) -> None:
        """
        Log a confusion matrix to wandb if logging to wandb is enabled.

        A chart object is created from the confusion matrix data and logged
        to the wandb run.

        Parameters
        ----------
        task_prefix: str
            The task's name, to be used in the matrix name.
        confusion_matrix : List[List[int]]
            the confusion matrix values
        labels : List[str]
            label names
        """
        if self.wandb_run:
            conf_matrix_data = []
            for row_idx, row in enumerate(confusion_matrix):
                for col_idx, cell_value in enumerate(row):
                    conf_matrix_data.append([labels[row_idx], labels[col_idx], cell_value])
            table = wandb.Table(columns=["Predicted", "Actual", "Count"], data=conf_matrix_data)
            chart = wandb.visualize("wandb/confusion_matrix/v1", table)
            self.wandb_run.log({f"{task_prefix}/confusion_matrix": chart})

    def log_to_summary(self, task_prefix, metric_name, metric_value) -> None:
        """
        Add a metric to the W&B run summary if logging to W&B is enabled.

        Parameters
        ----------
        task_prefix : str
            task name to be used as section in the summary table
        metric_name : str
            The metric name
        metric_value : Any
            The metric value
        """
        if self.wandb_run:
            self.wandb_run.summary[f"{task_prefix}/{metric_name}"] = metric_value


def get_config_dict(config_file_path: str) -> Dict[str, Any]:
    """
    Load a configuration file into a dictionary, to be logged to W&B as a run config.

    Parameters
    ----------
    config_file_path : str
        Path to the config file

    Returns
    -------
    Dictionary containing all SKLL configuration fields.

    This also includes default values when for fields that are missing in the file.
    """
    config_parser = _setup_config_parser(config_file_path, validate=False)
    return {
        section: {key: val for [key, val] in config_parser.items(section)}
        for section in config_parser.sections()
    }



# License: BSD 3 clause
"""
Utility classes and functions to parse SKLL configuration files.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
"""

import csv
import errno
import logging
from pathlib import Path
from typing import Iterable, List, Union

from ruamel.yaml import YAML

from skll.types import FoldMapping, PathOrStr


def fix_json(json_string: str) -> str:
    """
    Fix incorrectly formatted quotes and capitalized booleans in JSON string.

    Parameters
    ----------
    json_string : str
        A JSON-style string.

    Returns
    -------
    str
        The normalized JSON string.
    """
    json_string = json_string.replace("True", "true")
    json_string = json_string.replace("False", "false")
    json_string = json_string.replace("'", '"')
    return json_string


def load_cv_folds(folds_file: PathOrStr, ids_to_floats=False) -> FoldMapping:
    """
    Load cross-validation folds from a CSV file.

    The CSV file must contain two columns: example ID and fold ID (and a header).

    Parameters
    ----------
    folds_file : :class:`skll.types.PathOrStr`
        The path to a folds file to read.

    ids_to_floats : bool, default=False
        Whether to convert IDs to floats.

    Returns
    -------
    :class:`skll.types.FoldMapping`
        Dictionary with example IDs as the keys and fold IDs as the values.
        If `ids_to_floats` is set to `True`, the example IDs are floats but
        otherwise they are strings.

    Raises
    ------
    ValueError
        If example IDs cannot be converted to floats and `ids_to_floats` is `True`.
    """
    with open(folds_file) as f:
        reader = csv.reader(f)
        next(reader)  # discard the header
        res = {}
        for row in reader:
            example_id: Union[float, str] = row[0]
            fold_id: str = row[1]
            if ids_to_floats:
                try:
                    example_id = float(example_id)
                except ValueError:
                    raise ValueError(
                        "You set ids_to_floats to true, but ID "
                        f"{row[0]} could not be converted to "
                        "float"
                    )
            res[example_id] = fold_id

    return res


def locate_file(file_path: PathOrStr, config_dir: PathOrStr) -> str:
    """
    Locate a file, given a file path and configuration directory.

    Parameters
    ----------
    file_path : :class:`skll.types.PathOrStr`
        The file to locate. Path may be absolute or relative.

    config_dir : :class:`skll.types.PathOrStr`
        The path to the configuration file directory.

    Returns
    -------
    path_to_check : str
        The normalized absolute path, if it exists.

    Raises
    ------
    FileNotFoundError
        If the file does not exist.
    """
    if not file_path:
        return ""

    # convert to Path objects
    file_path = Path(file_path)
    config_dir = Path(config_dir)

    config_relative_path = config_dir / file_path
    path_to_check = file_path if file_path.is_absolute() else config_relative_path.resolve()
    if not path_to_check.exists():
        raise FileNotFoundError(
            errno.ENOENT, f"File {path_to_check} does not exist", str(path_to_check)
        )
    else:
        return str(path_to_check)


def _munge_featureset_name(name_or_list: Union[Iterable, str]) -> str:
    """
    Create a munged name for the featureset.

    Join features in ``featureset`` with a '+' if ``featureset`` is not a string.
    Otherwise, returns ``featureset``.

    Parameters
    ----------
    name_or_list : Union[Iterable, str]
        A featureset name or name components in a list.

    Returns
    -------
    res : str
        name components joined with '+' if input is a list or the name itself.
    """
    if isinstance(name_or_list, str):
        return name_or_list

    res = "+".join(sorted(name_or_list))
    return res


def _parse_and_validate_metrics(metrics: str, option_name: str, logger=None) -> List[str]:
    """
    Parse and validate string containing list of metrics.

    Given a string containing a list of metrics, this function
    parses that string into a list and validates some specific
    metric names.

    Parameters
    ----------
    metrics : str
        A string containing a list of metrics.

    option_name : str
        The name of the option with which the metrics are associated.

    logger : Optional[logging.Logger], default=None
        A logging object.

    Returns
    -------
    metrics : List[str]
        A list of metrics for the given option.

    Raises
    ------
    TypeError
        If the given string cannot be converted to a list.

    ValueError
        If "mean_squared_error" is specified as a metric.
    """
    # create a logger if one was not passed in
    if not logger:
        logger = logging.getLogger(__name__)

    # make sure the given metrics data type is a list
    # and parse it correctly
    yaml = YAML(typ="safe", pure=True)
    metrics = yaml.load(fix_json(metrics))
    if not isinstance(metrics, list):
        raise TypeError(f"{option_name} should be a list, not a " f"{type(metrics)}.")

    # `mean_squared_error` is no longer supported.
    # It has been replaced by `neg_mean_squared_error`
    if "mean_squared_error" in metrics:
        raise ValueError(
            'The metric "mean_squared_error" is no longer '
            "supported. please use the metric "
            '"neg_mean_squared_error" instead.'
        )

    return metrics



# License: BSD 3 clause
"""
Functions related to running experiments and parsing configuration files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:author: Chee Wee Leong (cleong@ets.org)
"""

from __future__ import annotations

import csv
import json
import logging
import math
import sys
from collections import defaultdict
from pathlib import Path
from typing import IO, Any, Dict, List, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from ruamel.yaml import YAML

from skll.types import FoldMapping, PathOrStr
from skll.utils.logging import MatplotlibCategoryFilter, get_skll_logger

# Turn off interactive plotting for matplotlib
plt.ioff()


def _compute_ylimits_for_featureset(
    df: pd.DataFrame, metrics: List[str]
) -> Dict[str, Tuple[float, float]]:
    """
    Compute the y-limits for learning curve score plots.

    Parameters
    ----------
    df : pd.DataFrame
        A data_frame with relevant metric information for
        train and test.
    metrics : List[str]
        A list of metrics for learning curve plots.

    Returns
    -------
    ylimits : Dict[str, Tuple[float, float]]
        A dictionary, with metric names as keys
        and a tuple of (lower_limit, upper_limit) as values.
    """
    # set the y-limits of the curves depending on what kind
    # of values the metric produces
    ylimits = {}
    for metric in metrics:
        # get the real min and max for the values that will be plotted
        df_train = df[(df["variable"] == "train_score_mean") & (df["metric"] == metric)]
        df_test = df[(df["variable"] == "test_score_mean") & (df["metric"] == metric)]
        train_values_lower = df_train["value"].values - df_train["train_score_std"].values
        test_values_lower = df_test["value"].values - df_test["test_score_std"].values
        min_score = np.min(np.concatenate([train_values_lower, test_values_lower]))
        train_values_upper = df_train["value"].values + df_train["train_score_std"].values
        test_values_upper = df_test["value"].values + df_test["test_score_std"].values
        max_score = np.max(np.concatenate([train_values_upper, test_values_upper]))

        # squeeze the limits to hide unnecessary parts of the graph
        # set the limits with a little buffer on either side but not too much
        if min_score < 0:
            lower_limit = max(min_score - 0.1, math.floor(min_score) - 0.05)
        else:
            lower_limit = 0

        if max_score > 0:
            upper_limit = min(max_score + 0.1, math.ceil(max_score) + 0.05)
        else:
            upper_limit = 0

        ylimits[metric] = (lower_limit, upper_limit)

    return ylimits


def _generate_learning_curve_score_plots(
    df_scores: pd.DataFrame,
    num_metrics: int,
    num_learners: int,
    experiment_name: str,
    output_dir: PathOrStr,
    rotate_labels: bool = False,
) -> List[str]:
    """
    Generate learning curve score plots, one per featureset.

    This function generates score-based learning curve plots, i.e., plots
    where the training set size is on the x-axis and the training and
    cross-validation scores are on the y-axis. The plots are faceted, with
    different metrics along the rows and different learners along the columns.

    Parameters
    ----------
    df_scores : pandas.DataFrame
        The pandas data frame containing the various scores to be plotted.
    num_metrics : int
        The number of metrics specified in the experiment.
    num_learners: int
        The number of learners specified in the experiment.
    experiment_name : str
        The name of the experiment.
    output_dir : :class:`skll.types.PathOrStr`
        Path to the output directory for the plots.
    rotate_labels : bool, default=False
        Whether to rotate the x-axis labels for the training data size.

    Returns
    -------
    List[str]
        A list of paths of the generated plots
    """
    # convert output dir to a path
    output_dir = Path(output_dir)
    plot_paths = []
    # set up and draw the actual learning curve figures, one for
    # each of the featuresets
    for fs_name, df_fs in df_scores.groupby("featureset_name"):
        fig = plt.figure()
        fig.set_size_inches(2.5 * num_learners, 2.5 * num_metrics)

        # compute ylimits for this feature set for each objective
        with sns.axes_style("whitegrid", {"grid.linestyle": ":", "xtick.major.size": 3.0}):
            g = sns.FacetGrid(
                df_fs,
                row="metric",
                col="learner_name",
                height=2.5,
                aspect=1,
                margin_titles=True,
                despine=True,
                sharex=False,
                sharey=False,
                legend_out=False,
            )
            train_color, test_color = sns.color_palette(palette="Set1", n_colors=2)

            # create a filter to hide unnecessary matplotlib.category warnings
            filter = MatplotlibCategoryFilter()
            logging.getLogger("matplotlib.category").addFilter(filter)

            # map the FacetGrid to the data
            g = g.map_dataframe(
                sns.pointplot,
                x="training_set_size",
                y="value",
                hue="variable",
                markersize=2,
                linewidth=0.8,
                errorbar=None,
                palette={"train_score_mean": train_color, "test_score_mean": test_color},
            )
            ylimits = _compute_ylimits_for_featureset(df_fs, g.row_names)
            for ax in g.axes.flat:
                plt.setp(ax.texts, text="")
            g = g.set_titles(row_template="", col_template="{col_name}").set_axis_labels(
                "Training Examples", "Score"
            )
            if rotate_labels:
                g = g.set_xticklabels(rotation=60)

            for i, row_name in enumerate(g.row_names):
                for j, col_name in enumerate(g.col_names):
                    ax = g.axes[i][j]
                    ax.set(ylim=ylimits[row_name])
                    df_ax_train = df_fs[
                        (df_fs["learner_name"] == col_name)
                        & (df_fs["metric"] == row_name)
                        & (df_fs["variable"] == "train_score_mean")
                    ]
                    df_ax_test = df_fs[
                        (df_fs["learner_name"] == col_name)
                        & (df_fs["metric"] == row_name)
                        & (df_fs["variable"] == "test_score_mean")
                    ]
                    ax.fill_between(
                        list(range(len(df_ax_train))),
                        df_ax_train["value"] - df_ax_train["train_score_std"],
                        df_ax_train["value"] + df_ax_train["train_score_std"],
                        alpha=0.1,
                        color=train_color,
                    )
                    ax.fill_between(
                        list(range(len(df_ax_test))),
                        df_ax_test["value"] - df_ax_test["test_score_std"],
                        df_ax_test["value"] + df_ax_test["test_score_std"],
                        alpha=0.1,
                        color=test_color,
                    )
                    if j == 0:
                        ax.set_ylabel(row_name)
                        if i == 0:
                            # set up the legend handles for this plot
                            plot_handles = [
                                matplotlib.lines.Line2D(
                                    [], [], color=color, label=label, linestyle="-"
                                )
                                for color, label in zip(
                                    [train_color, test_color], ["Training", "Cross-validation"]
                                )
                            ]
                            ax.legend(
                                handles=plot_handles,
                                loc=4,
                                fancybox=True,
                                fontsize="x-small",
                                ncol=1,
                                frameon=True,
                            )
            g.figure.tight_layout(w_pad=1)
            plot_file_path = output_dir / f"{experiment_name}_{fs_name}.png"
            plt.savefig(plot_file_path, dpi=300)
            # explicitly close figure to save memory
            plt.close(fig)
            plot_paths.append(str(plot_file_path))

    return plot_paths


def _generate_learning_curve_time_plots(
    df_times: pd.DataFrame,
    num_learners: int,
    experiment_name: str,
    output_dir: PathOrStr,
    rotate_labels: bool = False,
) -> List[str]:
    """
    Generate learning curve time plots, one per featureset.

    This function generates time-based learning curve plots, i.e., plots
    where the training set size is on the x-axis and the model fit times
    are on the y-axis. The plots are faceted, with different learners
    along the columns. There is a single row.

    Parameters
    ----------
    df_times : pandas.DataFrame
        The pandas data frame containing the various fit times to be plotted.
    num_learners: int
        The number of learners specified in the experiment.
    experiment_name : str
        The name of the experiment.
    output_dir : :class:`skll.types.PathOrStr`
        Path to the output directory for the plots.
    rotate_labels : bool, default=False
        Whether to rotate the x-axis labels for the training data size.

    Returns
    -------
     List[str]
        A list of paths of the generated plots
    """
    # convert output dir to a path
    output_dir = Path(output_dir)
    plot_paths = []
    # set up and draw the actual learning curve figures, one for
    # each of the featuresets
    for fs_name, df_fs in df_times.groupby("featureset_name"):
        fig = plt.figure()
        fig.set_size_inches(2.5 * num_learners, 2.5)

        # compute ylimits for this feature set for each metric
        with sns.axes_style("whitegrid", {"grid.linestyle": ":", "xtick.major.size": 3.0}):
            g = sns.FacetGrid(
                df_fs,
                col="learner_name",
                height=2.5,
                aspect=1,
                margin_titles=True,
                despine=True,
                sharex=False,
                sharey=True,
                legend_out=False,
            )

            g = g.map_dataframe(
                sns.pointplot,
                x="training_set_size",
                y="value",
                hue="variable",
                markersize=2,
                linewidth=0.8,
                errorbar=None,
                palette="dark:#1f77b4",
            )
            # compute the upper and lower
            for ax in g.axes.flat:
                plt.setp(ax.texts, text="")
            g = g.set_titles(row_template="", col_template="{col_name}").set_axis_labels(
                "Training Examples", "Fit time (s)"
            )
            if rotate_labels:
                g = g.set_xticklabels(rotation=60)

            for j, col_name in enumerate(g.col_names):
                ax = g.axes[0][j]
                df_ax = df_fs[
                    (df_fs["learner_name"] == col_name) & (df_fs["variable"] == "fit_time_mean")
                ]
                ax.fill_between(
                    list(range(len(df_ax))),
                    df_ax["value"] - df_ax["fit_time_std"],
                    df_ax["value"] + df_ax["fit_time_std"],
                    alpha=0.1,
                )

            g.figure.tight_layout(w_pad=1)
            plot_file_path = output_dir / f"{experiment_name}_{fs_name}_times.png"
            plt.savefig(plot_file_path, dpi=300)

            # explicitly close figure to save memory
            plt.close(fig)
            plot_paths.append(str(plot_file_path))

    return plot_paths


def generate_learning_curve_plots(
    experiment_name: str, output_dir: PathOrStr, learning_curve_tsv_file: PathOrStr
) -> List[str]:
    """
    Generate learning curves using the TSV output file from a learning curve experiment.

    This function generates both the score plots as well as the fit time plots.

    Parameters
    ----------
    experiment_name : str
        The name of the experiment.
    output_dir : :class:`skll.types.PathOrStr`
        Path to the output directory for the plots.
    learning_curve_tsv_file : :class:`skll.types.PathOrStr`
        The path to the learning curve TSV file.

    Returns
    -------
    List[str]
        A list of paths of the generated plots
    """
    # convert output_dir to Path object
    output_dir = Path(output_dir)

    # use pandas to read in the TSV file into a data frame
    # and massage it from wide to long format for plotting
    df = pd.read_csv(learning_curve_tsv_file, sep="\t")
    num_learners = len(df["learner_name"].unique())
    num_metrics = len(df["metric"].unique())

    # if there are any training sizes greater than 1000,
    # then we should probably rotate the tick labels
    # since otherwise the labels are not clearly rendered
    rotate_labels = df["training_set_size"].unique().max() >= 1000

    # get the columns relevant to the two types of plots
    score_columns = [
        column
        for column in df.columns
        if column
        not in [
            "train_set_name",
            "fit_time_mean",
            "fit_time_std",
            "scikit_learn_version",
            "version",
        ]
    ]
    time_columns = [
        column
        for column in df.columns
        if column
        not in [
            "train_set_name",
            "train_score_mean",
            "train_score_std",
            "test_score_mean",
            "test_score_std",
            "scikit_learn_version",
            "version",
        ]
    ]

    # create the score-specific data frame
    df_score = df[score_columns].copy()
    df_score_melted = pd.melt(
        df_score,
        id_vars=[c for c in df_score.columns if c not in ["train_score_mean", "test_score_mean"]],
    )
    # make sure the "variable" column is categorical since it will be
    # mapped to hue levels in the learning curve below
    df_score_melted["variable"] = df_score_melted["variable"].astype("category")

    # also make sure that the "learner_name" column is categorical so that
    # it's sorted correctly
    df_score_melted["learner_name"] = df_score_melted["learner_name"].astype("category")

    # now compute the time-specific data frame
    df_time = df[time_columns].copy()

    # note that although we have already averaged the fit times over
    # the various training, we still have multiple fit times for each
    # of the metrics so we can further average those out
    df_time_indexed = df_time.set_index(
        ["featureset_name", "learner_name", "training_set_size", "metric"]
    )
    df_time = (
        df_time_indexed.groupby(level=["featureset_name", "learner_name", "training_set_size"])
        .mean()
        .reset_index()
    )

    # now let's melt the time data frame the same way that we did the score one
    df_time_melted = pd.melt(
        df_time,
        id_vars=[c for c in df_time.columns if c != "fit_time_mean"],
    )

    # also make sure that the "learner_name" column is categorical so that
    # it's sorted correctly
    df_time_melted["learner_name"] = df_time_melted["learner_name"].astype("category")

    # call the function to generate the score plots first
    score_plot_paths = _generate_learning_curve_score_plots(
        df_score_melted,
        num_metrics,
        num_learners,
        experiment_name,
        output_dir,
        rotate_labels=rotate_labels,
    )

    # now call the function to generate the time plots
    time_plot_paths = _generate_learning_curve_time_plots(
        df_time_melted,
        num_learners,
        experiment_name,
        output_dir,
        rotate_labels=rotate_labels,
    )

    return score_plot_paths + time_plot_paths


def _print_fancy_output(
    learner_result_dicts: List[Dict[str, Any]], output_file: IO[str] = sys.stdout
) -> None:
    """
    Print nice tables with all of the results from cross-validation folds.

    Parameters
    ----------
    learner_result_dicts : List[Dict[str, Any]]
        List of result dictionaries.
    output_file : IO[str], default=sys.stdout
        The file buffer to print to.
    """
    if not learner_result_dicts:
        raise ValueError("Result dictionary list is empty!")

    lrd = learner_result_dicts[0]
    print(f'Experiment Name: {lrd["experiment_name"]}', file=output_file)
    print(f'SKLL Version: {lrd["version"]}', file=output_file)
    print(f'Training Set: {lrd["train_set_name"]}', file=output_file)
    print(f'Training Set Size: {lrd["train_set_size"]}', file=output_file)
    print(f'Test Set: {lrd["test_set_name"]}', file=output_file)
    print(f'Test Set Size: {lrd["test_set_size"]}', file=output_file)
    print(f'Shuffle: {lrd["shuffle"]}', file=output_file)
    print(f'Feature Set: {lrd["featureset"]}', file=output_file)
    print(f'Learner: {lrd["learner_name"]}', file=output_file)
    print(f'Task: {lrd["task"]}', file=output_file)
    if lrd["folds_file"]:
        print(f'Specified Folds File: {lrd["folds_file"]}', file=output_file)
    if lrd["task"] == "cross_validate":
        print(f'Number of Folds: {lrd["cv_folds"]}', file=output_file)
        if not lrd["cv_folds"].endswith("folds file"):
            print(f'Stratified Folds: {lrd["stratified_folds"]}', file=output_file)
    print(f'Feature Scaling: {lrd["feature_scaling"]}', file=output_file)
    print(f'Grid Search: {lrd["grid_search"]}', file=output_file)
    if lrd["grid_search"]:
        print(f'Grid Search Folds: {lrd["grid_search_folds"]}', file=output_file)
        print(f'Grid Objective Function: {lrd["grid_objective"]}', file=output_file)
    if (
        lrd["task"] == "cross_validate"
        and lrd["grid_search"]
        and lrd["cv_folds"].endswith("folds file")
    ):
        print(
            "Using Folds File for Grid Search: " f'{lrd["use_folds_file_for_grid_search"]}',
            file=output_file,
        )
    if lrd["task"] in ["evaluate", "cross_validate"] and lrd["additional_scores"]:
        print(
            "Additional Evaluation Metrics: " f'{list(lrd["additional_scores"].keys())}',
            file=output_file,
        )
    print(f'Scikit-learn Version: {lrd["scikit_learn_version"]}', file=output_file)
    print(f'Start Timestamp: {lrd["start_timestamp"]}', file=output_file)
    print(f'End Timestamp: {lrd["end_timestamp"]}', file=output_file)
    print(f'Total Time: {lrd["total_time"]}', file=output_file)
    print("\n", file=output_file)

    for lrd in learner_result_dicts:
        if lrd["fold"]:  # otherwise it's a single results item (evaluate task)
            print(f'Fold: {lrd["fold"]}', file=output_file)
        print(f'Model Parameters: {lrd.get("model_params", "")}', file=output_file)
        print(f'Grid Objective Score (Train) = {lrd.get("grid_score", "")}', file=output_file)
        if "result_table" in lrd:
            print(lrd["result_table"], file=output_file)
            print(f'Accuracy = {lrd["accuracy"]}', file=output_file)
        if "descriptive" in lrd:
            print("Descriptive statistics:", file=output_file)
            for desc_stat in ["min", "max", "avg", "std"]:
                actual = lrd["descriptive"]["actual"][desc_stat]
                predicted = lrd["descriptive"]["predicted"][desc_stat]
                print(
                    f" {desc_stat.title()} = {actual:.4f} (actual), "
                    f"{predicted:.4f} (predicted)",
                    file=output_file,
                )
            print(f'Pearson = {lrd["pearson"]:f}', file=output_file)
        print(f'Objective Function Score (Test) = {lrd.get("score", "")}', file=output_file)

        # now print the additional metrics, if there were any
        if lrd["additional_scores"]:
            print("", file=output_file)
            print("Additional Evaluation Metrics (Test):", file=output_file)
            for metric, score in lrd["additional_scores"].items():
                score = "" if np.isnan(score) else score
                print(f" {metric} = {score}", file=output_file)
        print("", file=output_file)


def _write_learning_curve_file(result_json_paths: List[str], output_file: IO[str]) -> None:
    """
    Combine individual learning curve results JSON files into single TSV.

    Take a list of paths to individual learning curve results json files and
    write out a single TSV file with the learning curve data.

    Parameters
    ----------
    result_json_paths : List[str]
        list of paths to the individual result JSON files.
    output_file : IO[str]
        The file buffer to write to.
    """
    learner_result_dicts = []

    # Map from feature set names to all features in them
    logger = get_skll_logger("experiment")
    for json_path_str in result_json_paths:
        json_path = Path(json_path_str)
        if not json_path.exists():
            logger.error(
                f"JSON results file {json_path} not found. Skipping "
                "summary creation. You can manually create the "
                "summary file after the fact by using the "
                "summarize_results script."
            )
            return
        else:
            with open(json_path) as json_file:
                obj = json.load(json_file)
                learner_result_dicts.extend(obj)

    # Build and write header
    header = [
        "featureset_name",
        "learner_name",
        "metric",
        "train_set_name",
        "training_set_size",
        "train_score_mean",
        "test_score_mean",
        "fit_time_mean",
        "train_score_std",
        "test_score_std",
        "fit_time_std",
        "scikit_learn_version",
        "version",
    ]
    writer = csv.DictWriter(output_file, header, extrasaction="ignore", dialect=csv.excel_tab)
    writer.writeheader()

    # write out the fields we need for the learning curve file
    # specifically, we need to separate out the curve sizes
    # and scores into individual entries.
    for lrd in learner_result_dicts:
        training_set_sizes = lrd["computed_curve_train_sizes"]
        train_scores_means_by_size = lrd["learning_curve_train_scores_means"]
        test_scores_means_by_size = lrd["learning_curve_test_scores_means"]
        fit_times_means_by_size = lrd["learning_curve_fit_times_means"]
        train_scores_stds_by_size = lrd["learning_curve_train_scores_stds"]
        test_scores_stds_by_size = lrd["learning_curve_test_scores_stds"]
        fit_times_stds_by_size = lrd["learning_curve_fit_times_stds"]

        # rename `grid_objective` to `metric` since the latter name can be confusing
        lrd["metric"] = lrd["grid_objective"]

        for (
            size,
            train_score_mean,
            test_score_mean,
            fit_time_mean,
            train_score_std,
            test_score_std,
            fit_time_std,
        ) in zip(
            training_set_sizes,
            train_scores_means_by_size,
            test_scores_means_by_size,
            fit_times_means_by_size,
            train_scores_stds_by_size,
            test_scores_stds_by_size,
            fit_times_stds_by_size,
        ):
            lrd["training_set_size"] = size
            lrd["train_score_mean"] = train_score_mean
            lrd["test_score_mean"] = test_score_mean
            lrd["fit_time_mean"] = fit_time_mean
            lrd["train_score_std"] = train_score_std
            lrd["test_score_std"] = test_score_std
            lrd["fit_time_std"] = fit_time_std

            writer.writerow(lrd)

    output_file.flush()


def _write_skll_folds(skll_fold_ids: FoldMapping, skll_fold_ids_file: IO[str]) -> None:
    """
    Take a dictionary of id->test-fold-number and write it to a file.

    Parameters
    ----------
    skll_fold_ids : FoldMapping
        Dictionary with ids as keys and test-fold-numbers as values.
    skll_fold_ids_file : IO[str]
        An open file handler to write to.
    """
    f = csv.writer(skll_fold_ids_file)
    f.writerow(["id", "cv_test_fold"])
    for example_id in skll_fold_ids:
        f.writerow([example_id, skll_fold_ids[example_id]])

    skll_fold_ids_file.flush()


def _write_summary_file(result_json_paths: List[str], output_file: IO[str], ablation: int = 0):
    """
    Summarize individual result JSON files.

    Take a list of paths to individual result json files and return a single
    file that summarizes all of them.

    Parameters
    ----------
    result_json_paths : List[str]
        A list of paths to the individual result JSON files.
    output_file : IO[str]
        The file buffer to write to.
    ablation : int, default=0
        The number of features to remove when doing ablation experiment.
    """
    learner_result_dicts = []
    # Map from feature set names to all features in them
    all_features = defaultdict(set)
    logger = get_skll_logger("experiment")
    yaml = YAML(typ="safe", pure=True)

    for json_path_str in result_json_paths:
        json_path = Path(json_path_str)
        if not json_path.exists():
            logger.error(
                f"JSON results file {json_path} not found. Skipping "
                "summary creation. You can manually create the "
                "summary file after the fact by using the "
                "summarize_results script."
            )
            return
        else:
            with open(json_path) as json_file:
                obj = json.load(json_file)
                featureset_name = obj[0]["featureset_name"]
                if ablation != 0 and "_minus_" in featureset_name:
                    parent_set = featureset_name.split("_minus_", 1)[0]
                    all_features[parent_set].update(yaml.load(obj[0]["featureset"]))
                learner_result_dicts.extend(obj)

    # Build and write header
    unique_columns = set(learner_result_dicts[0].keys()) - {"result_table", "descriptive"}
    if ablation != 0:
        unique_columns.add("ablated_features")
    header = sorted(unique_columns)
    writer = csv.DictWriter(output_file, header, extrasaction="ignore", dialect=csv.excel_tab)
    writer.writeheader()

    # Build "ablated_features" list and fix some backward compatible things
    for lrd in learner_result_dicts:
        featureset_name = lrd["featureset_name"]
        if ablation != 0:
            parent_set = featureset_name.split("_minus_", 1)[0]
            ablated_features = all_features[parent_set].difference(yaml.load(lrd["featureset"]))
            lrd["ablated_features"] = ""
            if ablated_features:
                lrd["ablated_features"] = json.dumps(sorted(ablated_features))

        # write out the new learner dict with the readable fields
        writer.writerow(lrd)

    output_file.flush()



#!/usr/bin/env python
# License: BSD 3 clause
"""
Filter a given feature file to remove unwanted features, labels, or IDs.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional

from skll.data.readers import EXT_TO_READER, safe_float
from skll.data.writers import EXT_TO_WRITER, ARFFWriter, CSVWriter, TSVWriter
from skll.version import __version__


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and gets things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Takes an input feature file and removes any instances or"
        " features that do not match the specified patterns.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="infile",
        required=True,
        help="input feature file (ends in .arff, .csv, " ".jsonlines, .ndj, or .tsv)",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="outfile",
        required=True,
        help="output feature file (must have same extension " "as input file)",
    )
    parser.add_argument(
        "-f",
        "--feature",
        help="A feature in the feature file you would like to "
        "keep.  If unspecified, no features are removed.",
        nargs="*",
    )
    parser.add_argument(
        "-I",
        "--id",
        help="An instance ID in the feature file you would "
        "like to keep.  If unspecified, no instances are"
        " removed based on their IDs.",
        nargs="*",
    )
    parser.add_argument(
        "--id_col",
        help="Name of the column which contains the instance " "IDs in ARFF, CSV, or TSV files.",
        default="id",
    )
    parser.add_argument(
        "--inverse",
        help="Instead of keeping features and/or examples in " "lists, remove them.",
        action="store_true",
    )
    parser.add_argument(
        "-L",
        "--label",
        help="A label in the feature file you would like to "
        "keep.  If unspecified, no instances are removed"
        " based on their labels.",
        nargs="*",
    )
    parser.add_argument(
        "-l",
        "--label_col",
        help="Name of the column which contains the class "
        "labels in ARFF, CSV, or TSV files. For ARFF "
        "files, this must be the final column to count "
        "as the label.",
        default="y",
    )
    parser.add_argument(
        "-rb",
        "--replace_blanks_with",
        help="Specifies a new value with which to replace "
        "blank values in all columns in a CSV/TSV file. To "
        "replace blanks differently in each column, use "
        "the SKLL Reader API directly.",
        default=None,
    )
    parser.add_argument(
        "-db",
        "--drop_blanks",
        action="store_true",
        help="Drop all lines/rows in a CSV/TSV file that "
        "contain any blank values. If each row of the file "
        "contains a blank value, an exception will be raised.",
        default=False,
    )
    parser.add_argument(
        "-q", "--quiet", help='Suppress printing of "Loading..." messages.', action="store_true"
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")
    logger = logging.getLogger(__name__)

    # all extensions except .libsvm can be processed
    valid_extensions = {ext for ext in EXT_TO_READER if ext != ".libsvm"}

    # make sure the input file extension is one we can process
    input_extension = Path(args.infile).suffix.lower()
    output_extension = Path(args.outfile).suffix.lower()

    if input_extension == ".libsvm":
        logger.error(
            "Cannot filter LibSVM files.  Please use skll_convert to"
            " convert to a different datatype first."
        )
        sys.exit(1)

    if input_extension not in valid_extensions:
        logger.error(
            "Input file must be in either .arff, .csv, .jsonlines, "
            ".ndj, or .tsv format. You specified: "
            f"{input_extension}"
        )
        sys.exit(1)

    if output_extension != input_extension:
        logger.error(
            "Output file must be in the same format as the input "
            f"file.  You specified: {output_extension}"
        )
        sys.exit(1)

    if input_extension == ".csv" or input_extension == ".tsv":
        replace_blanks_with = args.replace_blanks_with
        drop_blanks = args.drop_blanks
        if drop_blanks and replace_blanks_with is not None:
            raise ValueError(
                "You cannot both drop blanks and replace them. "
                "'replace_blanks_with' can only have a value "
                "when 'drop_blanks' is `False`."
            )
        replace_blanks_with = (
            None if replace_blanks_with is None else safe_float(replace_blanks_with)
        )
        kwargs = {"replace_blanks_with": replace_blanks_with, "drop_blanks": drop_blanks}
    else:
        kwargs = {}

    # Read input file
    reader = EXT_TO_READER[input_extension](
        args.infile, quiet=args.quiet, label_col=args.label_col, id_col=args.id_col, **kwargs
    )
    feature_set = reader.read()

    # convert the specified labels with `safe_float` before using
    # to match what the various featureset readers do
    converted_labels = None
    if args.label:
        converted_labels = [safe_float(label) for label in args.label]

    # Do the actual filtering
    feature_set.filter(
        ids=args.id, labels=converted_labels, features=args.feature, inverse=args.inverse
    )

    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[input_extension]
    writer_args = {"quiet": args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args["label_col"] = args.label_col
        writer_args["id_col"] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args["label_col"] = args.label_col
        writer_args["id_col"] = args.id_col
        writer_args["regression"] = reader.regression
        writer_args["relation"] = reader.relation
    writer = writer_type(args.outfile, feature_set, **writer_args)
    writer.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# License: BSD 3 clause
"""
Create a summary file from a list of JSON results files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import argparse
import logging
from typing import List, Optional

from skll.experiments.output import _write_summary_file
from skll.version import __version__


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and get things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Creates an experiment summary TSV file from a list of "
        "JSON files generated by run_experiment.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
    )
    parser.add_argument("summary_file", help="TSV file to store summary of results.")
    parser.add_argument(
        "json_file", help="JSON results file generated by run_experiment.", nargs="+"
    )
    parser.add_argument(
        "-a",
        "--ablation",
        action="store_true",
        default=False,
        help="The results files are from an ablation run.",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")

    with open(args.summary_file, "w") as output_file:
        _write_summary_file(args.json_file, output_file, ablation=int(args.ablation))


if __name__ == "__main__":
    main()



# License: BSD 3 clause
"""
Handles reading and writing data from various types of data files.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import csv

from .featureset import FeatureSet
from .readers import (
    ARFFReader,
    CSVReader,
    LibSVMReader,
    NDJReader,
    Reader,
    TSVReader,
    safe_float,
)
from .writers import ARFFWriter, CSVWriter, LibSVMWriter, NDJWriter, TSVWriter, Writer

# Register dialect for handling ARFF files
csv.register_dialect(
    "arff",
    delimiter=",",
    quotechar="'",
    escapechar="\\",
    doublequote=False,
    lineterminator="\n",
    skipinitialspace=True,
)


__all__ = [
    "Reader",
    "safe_float",
    "FeatureSet",
    "ARFFReader",
    "CSVReader",
    "LibSVMReader",
    "NDJReader",
    "TSVReader",
    "ARFFWriter",
    "LibSVMWriter",
    "TSVWriter",
    "CSVWriter",
    "NDJWriter",
    "Writer",
]



# License: BSD 3 clause
"""
Custom type aliases for readability.

:author: Nitin Madnani (nmadnani@ets.org)
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, Iterator, List, Optional, Tuple, Union

import numpy as np
from scipy.sparse import csr_matrix
from typing_extensions import TypeAlias

# conditional imports to avoid circular dependency problem
if TYPE_CHECKING:
    import skll.data
    import skll.learner

# a class map that maps new labels (string)
# to list of old labels (list of string)
ClassMap: TypeAlias = Dict[str, List[str]]

# confusion matrix
ConfusionMatrix: TypeAlias = List[List[int]]

# list of feature dictionaries
FeatureDict: TypeAlias = Dict[str, Any]
FeatureDictList: TypeAlias = List[FeatureDict]

# an iterator over two featuresets, usually test and train
FeaturesetIterator: TypeAlias = Iterator[Tuple["skll.data.FeatureSet", "skll.data.FeatureSet"]]

# a mapping from example ID to fold ID;
# the example ID may be a float or a str
FoldMapping: TypeAlias = Dict[Union[float, str], str]

# a float or a string; this is useful
# for SKLL IDs that can be both
IdType: TypeAlias = Union[float, str]

# generator over two index numpy arrays - usually train and test
IndexIterator: TypeAlias = Generator[Tuple[np.ndarray, np.ndarray], None, None]

# a float, int, or a string; this is useful
# for SKLL labels that can be both
LabelType: TypeAlias = Union[float, int, str]

# learning curve sizes can either be a numpy array or a list of floats or ints
LearningCurveSizes: TypeAlias = Union[
    List[Union[float, int]],
    np.ndarray,
]

# a generator that yields a three-tuple:
# - an example ID (float or str)
# - a label (int, float, or str)
# - a feature dictionary
FeatGenerator: TypeAlias = Generator[Tuple[IdType, Optional[LabelType], FeatureDict], None, None]

# a string path or Path object
PathOrStr: TypeAlias = Union[Path, str]

# a sparse matrix for features
SparseFeatureMatrix: TypeAlias = csr_matrix

# Learner evaluate task results 5-tuple:
# - confusion matrix (classifier) / None (regressor)
# - accuracy (classifier) / None (regressor)
# - dictionary of results
# - score for the grid objective, None if no grid search
# - dictionary of score for any additional metrics
ComputeEvalMetricsResults: TypeAlias = Tuple[
    Optional[ConfusionMatrix],
    Optional[float],
    Dict[LabelType, Any],
    Optional[float],
    Dict[str, Optional[float]],
]

# Learner evaluate task results 6-tuple:
# - confusion matrix (classifier) / None (regressor)
# - accuracy (classifier) / None (regressor)
# - dictionary of results
# - model parameters dictionary
# - score for the grid objective, None if no grid search
# - dictionary of score for any additional metrics
EvaluateTaskResults: TypeAlias = Tuple[
    Optional[ConfusionMatrix],
    Optional[float],
    Dict[LabelType, Any],
    Dict[str, Any],
    Optional[float],
    Dict[str, Optional[float]],
]

# Learner cross-validate task results 5-tuple:
# - the confusion matrix, overall accuracy, per-label PRFs, model parameters,
#   objective function score, and evaluation metrics (if any) for each fold.
# - the grid search scores for each fold.
# - list of dictionaries of grid search CV results, one per fold, with keys
#   such as "params", "mean_test_score", etc, that are mapped to lists of values
#   associated with each hyperparameter set combination.
# - dictionary containing the test-fold number for each, None if not saved
# - a list of learners, one for each fold, None if not saved
CrossValidateTaskResults: TypeAlias = Tuple[
    List[EvaluateTaskResults],
    List[float],
    List[Dict[str, Any]],
    Optional[FoldMapping],
    Optional[List["skll.learner.Learner"]],
]

# Voting Learner cross-validate task results 3-tuple:
# - the confusion matrix, overall accuracy, per-label PRFs, model parameters,
#   objective function score, and evaluation metrics (if any) for each fold.
# - the grid search scores for each fold.
# - dictionary containing the test-fold number for each, None if not saved
# - a list of learners, one for each fold, None if not saved
VotingCrossValidateTaskResults: TypeAlias = Tuple[
    List[EvaluateTaskResults],
    Optional[FoldMapping],
    Optional[List["skll.learner.voting.VotingLearner"]],
]



# License: BSD 3 clause
"""
The main class and functions used to parse SKLL configuration files.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Chee Wee Leong (cleong@ets.org)
"""

import configparser
import itertools
import logging
import os
from numbers import Number
from os.path import basename, join
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
from ruamel.yaml import YAML

from skll.data.readers import safe_float
from skll.types import ClassMap, FoldMapping, LabelType, PathOrStr
from skll.utils.constants import (
    PROBABILISTIC_METRICS,
    VALID_FEATURE_SCALING_OPTIONS,
    VALID_SAMPLERS,
    VALID_TASKS,
)
from skll.utils.logging import get_skll_logger

from .utils import (
    _munge_featureset_name,
    _parse_and_validate_metrics,
    fix_json,
    load_cv_folds,
    locate_file,
)

__all__ = ["SKLLConfigParser", "fix_json", "load_cv_folds", "locate_file"]


class SKLLConfigParser(configparser.ConfigParser):
    """A custom configuration file parser for SKLL."""

    def __init__(self) -> None:
        """Initialize configuration parser."""
        # these are the three options that must be set in a config
        # file and no defaults are provided
        required = ["experiment_name", "task", "learners"]

        # these are the optional config options for which
        # defaults are automatically provided
        defaults = {
            "class_map": "{}",
            "custom_learner_path": "",
            "custom_metric_path": "",
            "cv_seed": "123456789",
            "folds_file": "",
            "feature_hasher": "False",
            "feature_scaling": "none",
            "featuresets": "[]",
            "featureset_names": "[]",
            "fixed_parameters": "[]",
            "grid_search": "True",
            "grid_search_folds": "5",
            "grid_search_jobs": "0",
            "hasher_features": "0",
            "id_col": "id",
            "ids_to_floats": "False",
            "label_col": "y",
            "logs": "",
            "learning_curve_cv_folds_list": "[]",
            "learning_curve_train_sizes": "[]",
            "min_feature_count": "1",
            "models": "",
            "num_cv_folds": "10",
            "metrics": "[]",
            "objectives": "[]",
            "param_grids": "[]",
            "pos_label": "",
            "pipeline": "False",
            "predictions": "",
            "probability": "False",
            "random_folds": "False",
            "results": "",
            "sampler": "",
            "sampler_parameters": "{}",
            "save_cv_folds": "True",
            "save_cv_models": "False",
            "shuffle": "False",
            "suffix": "",
            "test_directory": "",
            "test_file": "",
            "train_directory": "",
            "train_file": "",
            "use_folds_file_for_grid_search": "True",
            "save_votes": "False",
            "wandb_credentials": "{}",
        }

        correct_section_mapping = {
            "class_map": "Input",
            "custom_learner_path": "Input",
            "custom_metric_path": "Input",
            "cv_seed": "Input",
            "folds_file": "Input",
            "feature_hasher": "Input",
            "feature_scaling": "Input",
            "featuresets": "Input",
            "featureset_names": "Input",
            "fixed_parameters": "Input",
            "grid_search": "Tuning",
            "grid_search_folds": "Tuning",
            "grid_search_jobs": "Tuning",
            "hasher_features": "Input",
            "id_col": "Input",
            "ids_to_floats": "Input",
            "label_col": "Input",
            "logs": "Output",
            "learning_curve_cv_folds_list": "Input",
            "learning_curve_train_sizes": "Input",
            "min_feature_count": "Tuning",
            "metrics": "Output",
            "models": "Output",
            "num_cv_folds": "Input",
            "objectives": "Tuning",
            "param_grids": "Tuning",
            "pos_label": "Tuning",
            "pipeline": "Output",
            "predictions": "Output",
            "probability": "Output",
            "random_folds": "Input",
            "results": "Output",
            "sampler": "Input",
            "sampler_parameters": "Input",
            "save_cv_folds": "Output",
            "save_cv_models": "Output",
            "shuffle": "Input",
            "suffix": "Input",
            "test_directory": "Input",
            "test_file": "Input",
            "train_directory": "Input",
            "train_file": "Input",
            "use_folds_file_for_grid_search": "Tuning",
            "save_votes": "Output",
            "wandb_credentials": "Output",
        }

        # make sure that the defaults dictionary and the
        # section mapping dictionary have the same keys
        assert defaults.keys() == correct_section_mapping.keys()

        super(SKLLConfigParser, self).__init__(defaults=defaults)
        self._required_options = required
        self._section_mapping = correct_section_mapping

    def _find_invalid_options(self) -> Set[str]:
        """
        Find the set of invalid options specified by the user.

        Returns
        -------
        invalid_options : Set[str]
            The set of invalid options specified by the user.
        """
        # compute a list of all the valid options
        valid_options = list(self.defaults().keys()) + self._required_options

        # get a set of all of the specified options
        specified_options = set(
            itertools.chain(*[self.options(section) for section in self.sections()])
        )

        # find any invalid options and return
        invalid_options = set(specified_options).difference(valid_options)
        return invalid_options

    def _find_ill_specified_options(
        self,
    ) -> Tuple[List[Tuple[str, str]], List[Tuple[str, List[str]]]]:
        """
        Find any ill-specified options in the configuration file.

        Make sure that all the options are specified in the appropriate sections
        and are not specified in multiple spaces. One way to do this is to
        basically get the values for all the optional config options from each
        of the sections and then look at the section that has a non-default
        (specified) value and if that's not the right section as indicated by
        our section mapping, then we need to raise an exception.

        Returns
        -------
        incorrectly_specified_options : List[Tuple[str, str]]
            List of incorrectly specified options. Each element in the list
            is a tuple with the first item in the tuple being the option name
            and the second element being the incorrect section name in which
            the option was specified.

        multiply_specified_options : List[Tuple[str, List[str]]]
            List of options specified more than once. Each element in the list
            is a tuple with the first item in the tuple being the option name
            and the second item being the list of section names in which the
            option was multiply specified.

        Notes
        -----
        This will NOT work if the user specifies the default value for the
        option but puts it in the wrong section. However, since specifying
        the default value for the option  does not result in running an
        experiment with unexpected settings, this is not really a major
        problem.
        """
        incorrectly_specified_options = []
        multiply_specified_options = []
        for option_name, default_value in self.defaults().items():
            used_sections = [
                section
                for section in self.sections()
                if self.get(section, option_name) != default_value
            ]
            # skip options that are not specified in any sections
            if len(used_sections) == 0:
                pass
            # find all options that are specified in multiple sections
            elif len(used_sections) > 1:
                multiply_specified_options.append((option_name, used_sections))
            # find all options that are specified in incorrect sections
            elif used_sections[0] != self._section_mapping[option_name]:
                incorrectly_specified_options.append((option_name, used_sections[0]))

        return (incorrectly_specified_options, multiply_specified_options)

    def validate(self) -> None:
        """
        Validate specified options.

        The following checks are used for validation::

            (a) no invalid options are specified
            (b) options are not specified in multiple sections
            (c) options are specified in the correct section

        Raises
        ------
        KeyError
            If configuration file contains unrecognized options.

        KeyError
            If any options are defined in multiple sections.

        KeyError
            If any options are not defined in the appropriate sections.
        """
        invalid_options = self._find_invalid_options()
        if invalid_options:
            raise KeyError(
                "Configuration file contains the following "
                f"unrecognized options: {list(invalid_options)}"
            )

        (
            incorrectly_specified_options,
            multiply_specified_options,
        ) = self._find_ill_specified_options()
        if multiply_specified_options:
            raise KeyError(
                "The following are defined in multiple sections: "
                f"{[t[0] for t in multiply_specified_options]}"
            )
        if incorrectly_specified_options:
            raise KeyError(
                "The following are not defined in the appropriate "
                f"sections: {[t[0] for t in incorrectly_specified_options]}"
            )


def parse_config_file(
    config_path: PathOrStr, log_level=logging.INFO
) -> Tuple[
    str,
    str,
    str,
    Dict[str, Any],
    bool,
    int,
    str,
    str,
    str,
    str,
    str,
    List[List[str]],
    bool,
    str,
    bool,
    List[str],
    bool,
    bool,
    str,
    Optional[LabelType],
    str,
    int,
    str,
    Optional[int],
    Optional[Union[int, FoldMapping]],
    Optional[Union[int, FoldMapping]],
    int,
    bool,
    bool,
    bool,
    bool,
    List[Dict[str, Any]],
    List[Dict[str, Any]],
    List[str],
    List[str],
    str,
    str,
    str,
    str,
    bool,
    Optional[ClassMap],
    str,
    str,
    List[int],
    List[Number],
    List[str],
    bool,
    Optional[Dict[str, str]],
]:
    """
    Parse a SKLL experiment configuration file with the given path.

    Log messages with the given log level (default: INFO).

    Parameters
    ----------
    config_path : :class:`skll.types.PathOrStr`
        The path to the configuration file.

    log_level : int, default=logging.INFO
        Determines which messages should be logged. You can either pass
        an integer or one of the corresponding ``logging`` parameters, such
        as ``logging.INFO`` or ``logging.WARNING``.

    Returns
    -------
    experiment_name : str
        A string used to identify this particular experiment configuration.
        When generating result summary files, this name helps prevent
        overwriting previous summaries.

    task : str
        The types of experiment we're trying to run (e.g. 'cross_validate').

    sampler : str
        The name of a sampler to perform non-linear transformations of the input.

    fixed_sampler_parameters : Dict[str, Any]
        A dictionary containing parameters you want to have fixed for the sampler.

    feature_hasher : bool
        If True, this enables a high-speed, low-memory vectorizer that uses
        feature hashing for converting feature dictionaries into NumPy arrays
        instead of using a DictVectorizer.

    hasher_features : int
        The number of features used by the FeatureHasher if the feature_hasher
        flag is enabled.

    id_col : str
        The column with IDs.

    label_col : str
        The column with labels.

    train_set_name : str
        The name of the training set.

    test_set_name : str
        The name of the test set.

    suffix : str
        The file format the training/test files are in.

    featuresets : List[[List[str]]
        A list of lists of prefixes for the files containing
        the features you would like to train/test on.

    do_shuffle : bool
        Whether to shuffle the data.

    model_path : str
        The path to the model file(s).

    do_grid_search : bool
        Whether to perform grid search.

    grid_objectives : List[str]
        A list of scoring functions to use for tuning.

    probability : bool
        Whether to output probabilities for each class.

    pipeline : bool
        Whether to include the `pipeline` attribute in the
        trained model. This will increase the size of the
        model file.

    results_path : str
        Path to store result files in.

    pos_label : Optional[LabelType]
        The label for the positive class in the binary classification setting.

    feature_scaling : str
        How to scale features (e.g. 'with_mean').

    min_feature_count : int
        The minimum number of examples for which the value of a
        feature must be nonzero to be included in the model.

    folds_file : str
        The path to the folds_file, if specified.

    grid_search_jobs : Optional[int]
        Number of folds to run in parallel when using grid search.

    grid_search_folds : Optional[Union[int, FoldMapping]]
        Folds to use for grid search. This may be:
        - `None`: if no grid search is to be done
        - `int`: an integer specifying how many folds to use
        - `FoldMapping`: a dictionary mapping example IDs
          in the data to fold IDs. Example IDs may be either strings or
          floats depending on the value of `ids_to_floats`.

    cv_folds : Optional[Union[int, FoldMapping]]
        Folds to use for cross-validation. This may be:
        - `None`: if no cross-validation is to be done
        - `int`: an integer specifying how many folds to use
        - `FoldMapping`: a dictionary mapping example IDs
          in the data to fold IDs. Example IDs may be either strings or
          floats depending on the value of `ids_to_floats`.

    cv_seed : int
        The seed value for the random number generator used
        to create the cross-validation folds.

    save_cv_folds : bool
        Whether to save CV Folds to file.

    save_cv_models : bool
        Whether to save CV models.

    use_folds_file_for_grid_search : bool
        Whether to use folds file for grid search.

    do_stratified_folds : bool
        Whether to use random folds for cross-validation.

    fixed_parameter_list : List[Dict[str, Any]]
        List of dicts containing parameters you want to have fixed for
        each classifier in learners list.

    param_grid_list : List[Dict[str, Any]]
        List of parameter grids to search, one dict for each learner.

    featureset_names : List[str]
        The names of the featuresets used for each job.

    learners : List[str]
        A list of learners to try using.

    prediction_dir : str
        The directories where predictions are saved.

    log_path : str
        The path to the log file.

    train_path : str
        The path to a file containing feature to train on.

    test_path : str
        The path to a file containing features to test on.

    ids_to_floats : bool
        Whether to convert IDs to floats.

    class_map : Optional[ClassMap]
        A class map collapsing several labels into one. The keys
        are the collapsed labels and each key's value is the list of
        labels to be collapsed into said label.

    custom_learner_path : str
        Path to a .py file that defines a custom learner.

    custom_metric_path : str
        Path to a .py file that defines a custom metric.

    learning_curve_cv_folds_list : List[int]
        A list of integers specifying the number of folds to use for CV.

    learning_curve_train_sizes : List[Number]
        List of floats or integers representing relative or absolute numbers
        of training examples that will be used to generate the learning
        curve respectively.

    output_metrics : List[str]
        A list of output metrics to use.

    save_votes : bool
        Whether to save the individual predictions from voting learners.

    wandb_credentials : Dict[str,str]
        A dictionary holding W&B entity and project name, if logging to W&B is enabled.

    Raises
    ------
    IOError
        If configuration file name is empty

    ValueError
        If various configuration parameters are incorrectly specified,
        or cause conflicts.
    """
    # ensure that a path is specified
    if not config_path:
        raise FileNotFoundError("The name of the configuration file is empty.")

    # convert config_path to a Path object and get the full path
    config_path = Path(config_path).resolve()

    # ensure that the specified path actually exists
    if not config_path.exists():
        raise FileNotFoundError(f"The configuration file {config_path} does not exist.")

    # get the containing directory for the configuration file
    config_dir = config_path.parent

    # set up a config parser with the above default values
    config: SKLLConfigParser = _setup_config_parser(config_path)

    # extract parameters from the various sections in the config file

    ######################
    # 1. General section #
    ######################
    if config.has_option("General", "experiment_name"):
        experiment_name = config.get("General", "experiment_name")
    else:
        raise ValueError(
            "Configuration file does not contain experiment_name " "in the [General] section."
        )

    # next, get the log path before anything else since we need to
    # save all logging messages to a log file in addition to displaying
    # them on the console
    log_value: str = config.get("Output", "logs")

    try:
        log_path = locate_file(log_value, config_dir)
    except FileNotFoundError as e:
        log_path = e.filename
        os.makedirs(log_path)

    # Create a top-level log file under the log path
    main_log_file = join(log_path, f"{experiment_name}.log")

    # Now create a SKLL logger that will log to this file as well
    # as to the console. Use the log level provided - note that
    # we only have to do this the first time we call `get_skll_logger()`
    # with a given name.
    logger = get_skll_logger("experiment", filepath=main_log_file, log_level=log_level)

    if config.has_option("General", "task"):
        task = config.get("General", "task")
    else:
        raise ValueError("Configuration file does not contain task in the " "[General] section.")
    if task not in VALID_TASKS:
        raise ValueError(
            f"An invalid task was specified: {task}.  Valid " f'tasks are: {", ".join(VALID_TASKS)}'
        )

    ####################
    # 2. Input section #
    ####################
    sampler = config.get("Input", "sampler")
    if sampler not in VALID_SAMPLERS:
        raise ValueError(
            f"An invalid sampler was specified: {sampler}.  Valid"
            f' samplers are: {", ".join(VALID_SAMPLERS)}'
        )

    # produce warnings if feature_hasher is set but hasher_features
    # is less than or equal to zero.
    feature_hasher = config.getboolean("Input", "feature_hasher")
    hasher_features = config.getint("Input", "hasher_features")
    if feature_hasher:
        if hasher_features <= 0:
            raise ValueError(
                "Configuration file must specify a non-zero value "
                "for the option hasher_features when "
                "feature_hasher is True."
            )

    # produce warnings if hasher_features is set but feature_hasher
    # is not set correctly
    elif hasher_features > 0:
        logger.warning(
            "Ignoring hasher_features since feature_hasher is either" " missing or set to False."
        )

    if config.has_option("Input", "learners"):
        learners_string = config.get("Input", "learners")
    else:
        raise ValueError(
            "Configuration file does not contain list of learners " "in [Input] section."
        )

    yaml = YAML(typ="safe", pure=True)
    learners = yaml.load(fix_json(learners_string))

    if len(learners) == 0:
        raise ValueError(
            "Configuration file contains an empty list of learners" " in the [Input] section."
        )

    elif len(set(learners)) < len(learners):
        raise ValueError(
            "Configuration file contains the same learner multiple"
            " times, which is not currently supported.  Please use"
            " param_grids with tuning to find the optimal settings"
            " for the learner."
        )
    custom_learner_path = locate_file(config.get("Input", "custom_learner_path"), config_dir)

    # get the custom metric path, if specified, and locate it
    custom_metric_path = locate_file(config.get("Input", "custom_metric_path"), config_dir)

    # get the featuresets
    featuresets = yaml.load(config.get("Input", "featuresets"))

    # ensure that featuresets is either a list of features or a list of lists
    # of features
    if not isinstance(featuresets, list) or not all(isinstance(fs, list) for fs in featuresets):
        raise ValueError(
            "The featuresets parameter should be a list of "
            "features or a list of lists of features. You "
            f"specified: {featuresets}"
        )

    featureset_names = yaml.load(fix_json(config.get("Input", "featureset_names")))

    # ensure that featureset_names is a list of strings, if specified
    if featureset_names:
        if not isinstance(featureset_names, list) or not all(
            [isinstance(fs, str) for fs in featureset_names]
        ):
            raise ValueError(
                "The featureset_names parameter should be a list "
                f"of strings. You specified: {featureset_names}"
            )

    # get the value for learning_curve_cv_folds and ensure
    # that it's a list of the same length as the value of
    # learners. If it's not specified, then we just assume
    # that we are using 10 folds for each learner.
    learning_curve_cv_folds_list_string = config.get("Input", "learning_curve_cv_folds_list")
    learning_curve_cv_folds_list = yaml.load(fix_json(learning_curve_cv_folds_list_string))
    if len(learning_curve_cv_folds_list) == 0:
        learning_curve_cv_folds_list = [10] * len(learners)
    else:
        if (
            not isinstance(learning_curve_cv_folds_list, list)
            or not all([isinstance(fold, int) for fold in learning_curve_cv_folds_list])
            or not len(learning_curve_cv_folds_list) == len(learners)
        ):
            raise ValueError(
                "The learning_curve_cv_folds parameter should "
                "be a list of integers of the same length as "
                "the number of learners. You specified: "
                f"{learning_curve_cv_folds_list}"
            )

    # get the value for learning_curve_train_sizes and ensure
    # that it's a list of either integers (sizes) or
    # floats (proportions). If it's not specified, then we just
    # assume that we are using np.linspace(0.1, 1.0, 5).
    learning_curve_train_sizes_string = config.get("Input", "learning_curve_train_sizes")
    learning_curve_train_sizes = yaml.load(fix_json(learning_curve_train_sizes_string))
    if len(learning_curve_train_sizes) == 0:
        learning_curve_train_sizes = np.linspace(0.1, 1.0, 5).tolist()
    else:
        if not isinstance(learning_curve_train_sizes, list) or not all(
            [
                isinstance(size, int) or isinstance(size, float)
                for size in learning_curve_train_sizes
            ]
        ):
            raise ValueError(
                "The learning_curve_train_sizes parameter should "
                "be a list of integers or floats. You specified: "
                f"{learning_curve_train_sizes}"
            )

    # do we need to shuffle the training data
    do_shuffle = config.getboolean("Input", "shuffle")

    fixed_parameter_list = yaml.load(fix_json(config.get("Input", "fixed_parameters")))
    fixed_sampler_parameters = yaml.load(fix_json(config.get("Input", "sampler_parameters")))
    param_grid_list = yaml.load(fix_json(config.get("Tuning", "param_grids")))

    # read and normalize the value of `pos_label`
    pos_label_string = safe_float(config.get("Tuning", "pos_label"))
    pos_label: Optional[LabelType] = pos_label_string if pos_label_string else None

    # ensure that feature_scaling is specified only as one of the
    # four available choices
    feature_scaling = config.get("Input", "feature_scaling")
    if feature_scaling not in VALID_FEATURE_SCALING_OPTIONS:
        raise ValueError("Invalid value for feature_scaling parameter: " f"{feature_scaling}")

    suffix = config.get("Input", "suffix")
    label_col = config.get("Input", "label_col")
    id_col = config.get("Input", "id_col")
    ids_to_floats = config.getboolean("Input", "ids_to_floats")

    # read in the cross-validation seed
    cv_seed = config.getint("Input", "cv_seed")

    # if an external folds file is specified, then read it into a dictionary
    folds_file = locate_file(config.get("Input", "folds_file"), config_dir)
    num_cv_folds = config.getint("Input", "num_cv_folds")
    specified_folds_mapping = None
    specified_num_folds = None
    if folds_file:
        specified_folds_mapping = load_cv_folds(folds_file, ids_to_floats=ids_to_floats)
    else:
        # if no file is specified, then set the number of folds for cross-validation
        specified_num_folds = num_cv_folds if num_cv_folds else 10

    # whether or not to save the cv fold ids/models
    save_cv_folds = config.getboolean("Output", "save_cv_folds")
    save_cv_models = config.getboolean("Output", "save_cv_models")

    # whether or not to do stratified cross validation
    random_folds = config.getboolean("Input", "random_folds")
    if random_folds:
        if folds_file:
            logger.warning('Specifying "folds_file" overrides "random_folds".')
        do_stratified_folds = False
    else:
        do_stratified_folds = True

    # get all the input paths and directories (without trailing slashes)
    train_path = config.get("Input", "train_directory").rstrip(os.sep)
    test_path = config.get("Input", "test_directory").rstrip(os.sep)
    train_file = config.get("Input", "train_file")
    test_file = config.get("Input", "test_file")

    # make sure that featuresets is not an empty list unless
    # train_file and test_file are specified
    if (
        not train_file
        and not test_file
        and (isinstance(featuresets, list) and len(featuresets) == 0)
    ):
        raise ValueError("The 'featuresets' parameters cannot be an empty list.")

    # The user must specify either train_file or train_path, not both.
    if not train_file and not train_path:
        raise ValueError(
            'Invalid [Input] parameters: either "train_file" or '
            '"train_directory" must be specified in the '
            "configuration file."
        )

    # Either train_file or train_path must be specified.
    if train_file and train_path:
        raise ValueError(
            'Invalid [Input] parameters: only either "train_file"'
            ' or "train_directory" can be specified in the '
            "configuration file, not both."
        )

    # Cannot specify both test_file and test_path
    if test_file and test_path:
        raise ValueError(
            'Invalid [Input] parameters: only either "test_file" '
            'or "test_directory" can be specified in the '
            "configuration file, not both."
        )

    # if train_file is specified, then assign its value to train_path
    # this is a workaround to make this simple use case (a single train and
    # test file) compatible with the existing architecture using
    # featuresets
    if train_file:
        train_path = train_file
        featuresets = [[f"train_{basename(train_file)}"]]
        suffix = ""

    # if test_file is specified, then assign its value to test_path to
    # enable compatibility with the pre-existing featuresets architecture
    if test_file:
        test_path = test_file
        featuresets[0][0] += f"_test_{basename(test_file)}"

    # make sure all the specified paths/files exist
    train_path = locate_file(train_path, config_dir)
    test_path = locate_file(test_path, config_dir)

    # Get class mapping dictionary if specified
    class_map_string = config.get("Input", "class_map")
    yaml = YAML(typ="safe", pure=True)
    original_class_map = yaml.load(fix_json(class_map_string))
    if original_class_map:
        # Change class_map to map from originals to replacements instead of
        # from replacement to list of originals
        class_map = {}
        for replacement, original_list in original_class_map.items():
            for original in original_list:
                class_map[original] = replacement
        del original_class_map
    else:
        class_map = None

    #####################
    # 3. Output section #
    #####################
    probability = config.getboolean("Output", "probability")
    pipeline = config.getboolean("Output", "pipeline")

    # do we want to keep the predictions?
    # make sure the predictions path exists and if not create it
    try:
        prediction_dir = locate_file(config.get("Output", "predictions"), config_dir)
    except FileNotFoundError as e:
        prediction_dir = e.filename
        os.makedirs(prediction_dir)

    # make sure model path exists and if not, create it
    try:
        model_path = locate_file(config.get("Output", "models"), config_dir)
    except FileNotFoundError as e:
        model_path = e.filename
        os.makedirs(model_path)

    # make sure results path exists
    try:
        results_path = locate_file(config.get("Output", "results"), config_dir)
    except FileNotFoundError as e:
        results_path = e.filename
        os.makedirs(results_path)

    # what are the output metrics?
    output_metrics_string = config.get("Output", "metrics")
    output_metrics: List[str] = _parse_and_validate_metrics(
        output_metrics_string, "metrics", logger=logger
    )

    # do we want to save the individual predictions from voting
    # learner estimators?
    save_votes = config.getboolean("Output", "save_votes")

    # get wandb credentials if specified. Both entity and project must
    # be specified to enable W&B logging
    wandb_credentials = yaml.load(fix_json(config.get("Output", "wandb_credentials")))
    if wandb_credentials:
        if "wandb_entity" not in wandb_credentials or "wandb_project" not in wandb_credentials:
            logger.warning(
                "Logging to W&B not enabled "
                "since either `wandb_entity` or `wandb_project` is not specified!"
            )
            wandb_credentials = {}

    #####################
    # 4. Tuning section #
    #####################

    # do we need to run a grid search for the hyperparameters or are we just
    # using the defaults?
    do_grid_search = config.getboolean("Tuning", "grid_search")

    # parse any provided grid objective functions
    grid_objectives_string = config.get("Tuning", "objectives")
    grid_objectives: List[str] = _parse_and_validate_metrics(
        grid_objectives_string, "objectives", logger=logger
    )

    # if we are doing learning curves , we don't care about
    # grid search
    if task == "learning_curve" and do_grid_search:
        do_grid_search = False
        logger.warning(
            "Grid search is not supported during " "learning curve generation. Disabling."
        )

    # Check if `param_grids` is specified, but `do_grid_search` is False
    if param_grid_list and not do_grid_search:
        logger.warning(
            'Since "grid_search" is set to False, the specified' ' "param_grids" will be ignored.'
        )

    # Warn user about potential conflicts between parameter values
    # specified in `fixed_parameter_list` and values specified in
    # `param_grid_list` (or values passed in by default) if
    # `do_grid_search` is True
    if do_grid_search and fixed_parameter_list:
        logger.warning(
            'Note that "grid_search" is set to True and '
            '"fixed_parameters" is also specified. If there '
            "is a conflict between the grid search parameter"
            " space and the fixed parameter values, the "
            "fixed parameter values will take precedence."
        )

    # minimum number of examples a feature must be nonzero in to be included
    min_feature_count = config.getint("Tuning", "min_feature_count")

    # if an external folds file was specified do we use the same folds file
    # for the inner grid-search in cross-validate as well?
    use_folds_file_for_grid_search = config.getboolean("Tuning", "use_folds_file_for_grid_search")

    # how many jobs should we run in parallel for grid search
    grid_search_jobs: Optional[int] = config.getint("Tuning", "grid_search_jobs")
    if not grid_search_jobs:
        grid_search_jobs = None

    # how many folds should we run in parallel for grid search
    # the folds can either be a None, or number, or a mapping
    # from example IDs to fold IDs; we initialize to the specified
    # number to start with
    grid_search_folds: Optional[Union[int, FoldMapping]] = config.getint(
        "Tuning", "grid_search_folds"
    )

    # check whether the right things are set for the given task
    if (task == "evaluate" or task == "predict") and not test_path:
        raise ValueError("The test set must be set when task is evaluate or " "predict.")
    if task in ["cross_validate", "evaluate", "train"]:
        if do_grid_search and len(grid_objectives) == 0:
            raise ValueError(
                "Grid search is on. Either specify a list of tuning "
                "objectives or set `grid_search` to `false` in the "
                "Tuning section."
            )
        if not do_grid_search and len(grid_objectives) > 0:
            logger.warning(
                'Since "grid_search" is set to False, any specified'
                ' "objectives" will be ignored.'
            )
            grid_objectives = []
    if task in ["cross_validate", "train", "learning_curve"] and test_path:
        raise ValueError("The test set should not be set when task is " f"{task}.")
    if task in ["train", "predict"] and results_path and not do_grid_search:
        raise ValueError(
            "The results path should not be set when task is "
            f'{task} and "grid_search" is set to False.'
        )
    if task == "train" and not model_path:
        raise ValueError("The model path should be set when task is train.")
    if task in ["learning_curve", "train"] and prediction_dir:
        raise ValueError("The predictions path should not be set when task is " f"{task}.")
    if task == "learning_curve" and model_path:
        raise ValueError("The models path should not be set when task is " "learning_curve.")
    if task == "learning_curve":
        if len(grid_objectives) > 0:
            raise ValueError(
                'The "objectives" option is no longer supported'
                ' for the "learning_curve" task. Please use the'
                ' "metrics" option in the [Output] section '
                "instead."
            )
        if len(output_metrics) == 0:
            raise ValueError(
                'The "metrics" option must be set when ' 'the task is "learning_curve".'
            )

    # if any of the objectives or metrics require probabilities to be output,
    # probability must be specified as true
    specified_probabilistic_metrics = PROBABILISTIC_METRICS.intersection(
        grid_objectives + output_metrics
    )
    if specified_probabilistic_metrics and not probability:
        raise ValueError(
            "The 'probability' option must be 'true' "
            " to compute the following: "
            f"{list(specified_probabilistic_metrics)}."
        )

    # set the folds appropriately based on the task:
    #  (a) if the task is `train`/`evaluate`/`predict` and if an external
    #      fold mapping is specified then use that mapping for grid search
    #      instead of the value contained in `grid_search_folds`.
    #  (b) if the task is `cross_validate` and an external fold mapping is specified
    #      then use that mapping for the outer CV loop and for the inner grid-search
    #      loop. However, if  `use_folds_file_for_grid_search` is `False`, do not
    #      use the fold mapping for the inner loop.
    cv_folds = None
    if task in ["train", "evaluate", "predict"] and specified_folds_mapping:
        grid_search_folds = specified_folds_mapping
        # only print out the warning if the user actually wants to do grid search
        if do_grid_search:
            logger.warning(
                'Specifying "folds_file" overrides both '
                'explicit and default "grid_search_folds".'
            )
    if task == "cross_validate":
        cv_folds = specified_folds_mapping if specified_folds_mapping else specified_num_folds
        if specified_folds_mapping:
            logger.warning(
                'Specifying "folds_file" overrides both ' 'explicit and default "num_cv_folds".'
            )
            if use_folds_file_for_grid_search:
                grid_search_folds = cv_folds
            else:
                # only print out the warning if the user wants to do grid search
                if do_grid_search:
                    logger.warning(
                        'The specified "folds_file" will ' "not be used for inner grid search."
                    )
        if save_cv_models is True and not model_path:
            raise ValueError(
                "Output directory for models must be set if " '"save_cv_models" is set to true.'
            )

    # Create feature set names if unspecified
    if not featureset_names:
        featureset_names = [_munge_featureset_name(x) for x in featuresets]
    if len(featureset_names) != len(featuresets):
        raise ValueError(
            "Number of feature set names "
            f"({len(featureset_names)}) does not match number of"
            f" feature sets ({len(featuresets)})."
        )

    # store training/test set names for later use
    train_set_name = basename(train_path)
    test_set_name = basename(test_path) if test_path else "cv"

    return (
        experiment_name,
        task,
        sampler,
        fixed_sampler_parameters,
        feature_hasher,
        hasher_features,
        id_col,
        label_col,
        train_set_name,
        test_set_name,
        suffix,
        featuresets,
        do_shuffle,
        model_path,
        do_grid_search,
        grid_objectives,
        probability,
        pipeline,
        results_path,
        pos_label,
        feature_scaling,
        min_feature_count,
        folds_file,
        grid_search_jobs,
        grid_search_folds,
        cv_folds,
        cv_seed,
        save_cv_folds,
        save_cv_models,
        use_folds_file_for_grid_search,
        do_stratified_folds,
        fixed_parameter_list,
        param_grid_list,
        featureset_names,
        learners,
        prediction_dir,
        log_path,
        train_path,
        test_path,
        ids_to_floats,
        class_map,
        custom_learner_path,
        custom_metric_path,
        learning_curve_cv_folds_list,
        learning_curve_train_sizes,
        output_metrics,
        save_votes,
        wandb_credentials,
    )


def _setup_config_parser(config_path: PathOrStr, validate=True) -> SKLLConfigParser:
    """
    Return a config parser at a given path.

    Only implemented as a separate function to simplify testing.

    Parameters
    ----------
    config_path : :class:`skll.types.PathOrStr`
        The path to the configuration file.

    validate : bool, default=True
        Whether to validate the configuration file.

    Returns
    -------
    config : SKLLConfigParser
        A SKLL configuration object.

    Raises
    ------
    FileNotFoundError
        If the configuration file does not exist.
    """
    # initialize config parser with the given defaults
    config = SKLLConfigParser()

    # convert config_path to a Path object
    config_path = Path(config_path)

    # Read file if it exists
    if not config_path.exists():
        raise FileNotFoundError(f"The configuration file {config_path} does not exist.")

    config.read(config_path)

    if validate:
        config.validate()

    return config



# License: BSD 3 clause
"""
Classes related to storing/merging feature sets.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
:author: Jeremy Biggs (jbiggs@ets.org)
:organization: ETS
"""

from copy import deepcopy
from typing import Collection, List, Optional, Tuple, Union

import numpy as np
import scipy.sparse as sp
from pandas import DataFrame
from sklearn.feature_extraction import DictVectorizer, FeatureHasher

from skll.data.dict_vectorizer import DictVectorizer as SkllDictVectorizer
from skll.types import FeatGenerator, FeatureDictList, IdType, LabelType, SparseFeatureMatrix


class FeatureSet(object):
    """
    Encapsulate features, labels, and metadata for a given dataset.

    Parameters
    ----------
    name : str
        The name of this feature set.

    ids : Union[List[str], numpy.ndarray]
        Example IDs for this set.

    labels : Optional[Union[List[str], numpy.ndarray], default=None
        Labels for this set.

    features : Optional[Union[:class:`skll.types.FeatureDictList`, :class:`numpy.ndarray`]], default=None
        The features for each instance represented as either a
        list of dictionaries or a numpy array (if ``vectorizer`` is
        also specified).

    vectorizer : Optional[Union[:class:`sklearn.feature_extraction.DictVectorizer`, :class:`sklearn.feature_extraction.FeatureHasher`], default=None
        Vectorizer which will be used to generate the feature matrix.

    Warnings
    --------
    FeatureSets can only be equal if the order of the instances is
    identical because these are stored as lists/arrays. Since scikit-learn's
    ``DictVectorizer`` automatically sorts the underlying feature matrix
    if it is sparse, we do not do any sorting before checking for equality.
    This is not a problem because we _always_ use sparse matrices with
    ``DictVectorizer`` when creating FeatureSets.

    Notes
    -----
    If ids, labels, and/or features are not None, the number of rows in
    each array must be equal.
    """

    def __init__(
        self,
        name: str,
        ids: Union[List[str], np.ndarray],
        labels: Optional[Union[List[str], np.ndarray]] = None,
        features: Optional[Union[FeatureDictList, SparseFeatureMatrix]] = None,
        vectorizer: Optional[Union[DictVectorizer, FeatureHasher]] = None,
    ):
        """Initialize a FeatureSet instance."""
        super(FeatureSet, self).__init__()

        # clearly define the attribute types
        self.ids: np.ndarray
        self.labels: Optional[np.ndarray]
        self.features: Optional[SparseFeatureMatrix]
        self.vectorizer: Optional[Union[DictVectorizer, FeatureHasher]]

        self.name = name

        if isinstance(ids, list):
            self.ids = np.array(ids)
        elif isinstance(ids, np.ndarray):
            self.ids = ids
        else:
            raise ValueError("Ids must be a list or numpy array.")

        if isinstance(labels, list):
            labels = np.array(labels)
        self.labels = labels

        self.vectorizer = vectorizer

        # convert features from list of dictionaries to sparse array, if needed
        if isinstance(features, list):
            if self.vectorizer is None:
                self.vectorizer = SkllDictVectorizer(sparse=True)
            features_array: SparseFeatureMatrix = self.vectorizer.fit_transform(features)
            self.features = features_array
        else:
            self.features = features

        if self.features is not None:
            num_feats = self.features.shape[0]
            num_ids = self.ids.shape[0]
            if num_feats != num_ids:
                raise ValueError(
                    f"Number of IDs ({num_ids}) does not equal "
                    f"number of feature rows ({num_feats})"
                )
            if self.labels is None:
                self.labels = np.empty(num_feats)
                self.labels.fill(None)
            num_labels = self.labels.shape[0]
            if num_feats != num_labels:
                raise ValueError(
                    f"Number of labels ({num_labels}) does not "
                    f"equal number of feature rows ({num_feats})"
                )

    def __contains__(self, value):
        """
        Check if example ID is in the FeatureSet.

        Parameters
        ----------
        value
            The value to check.
        """
        return value in self.ids

    def __eq__(self, other):
        """
        Check whether two featuresets are the same.

        Parameters
        ----------
        other : :class:`skll.data.featureset.FeatureSet`
            The other ``FeatureSet`` to check equivalence with.

        Returns
        -------
        bool
            ``True`` if they are the same, ``False`` otherwise.

        Notes
        -----
        We consider feature values to be equal if any differences are in the
        sixth decimal place or higher.
        """
        return (
            self.ids.shape == other.ids.shape
            and self.labels.shape == other.labels.shape
            and self.features.shape == other.features.shape
            and (self.ids == other.ids).all()
            and (self.labels == other.labels).all()
            and np.allclose(self.features.data, other.features.data, rtol=1e-6)
            and (self.features.indices == other.features.indices).all()
            and (self.features.indptr == other.features.indptr).all()
            and self.vectorizer == other.vectorizer
        )

    def __iter__(self):
        """Iterate through (ID, label, feature_dict) tuples in feature set."""
        if self.features is not None:
            if not isinstance(self.vectorizer, DictVectorizer):
                raise ValueError(
                    "FeatureSets can only be iterated through if "
                    "they use a DictVectorizer for their feature "
                    "vectorizer."
                )
            for id_, label_, feats in zip(self.ids, self.labels, self.features):
                # reshape to a 2D matrix if we are not using a sparse matrix
                # to store the features
                feats = feats.reshape(1, -1) if not sp.issparse(feats) else feats

                # When calling inverse_transform we have to add [0] to get the
                # results for the current instance because it always returns a
                # 2D array
                yield (id_, label_, self.vectorizer.inverse_transform(feats)[0])
        else:
            return

    def __len__(self) -> int:
        """Return number of rows in the ``FeatureSet`` instance."""
        return self.features.shape[0] if self.features is not None else 0

    def __add__(self, other: "FeatureSet") -> "FeatureSet":
        """
        Combine two feature sets to create a new one.

        The combination is done assuming they both have the same instances
        with the same IDs in the same order.

        Parameters
        ----------
        other : :class:`skll.data.featureset.FeatureSet`
            The other ``FeatureSet`` to add to this one.

        Returns
        -------
        :class:`skll.data.featureset.FeatureSet
            The combined feature set.

        Raises
        ------
        ValueError
            If IDs are not in the same order in each ``FeatureSet`` instance.

        ValueError
            If either the 'features' or 'vectorizer' attributes are
            ``None`` for either of the two ``FeatureSet`` instances.

        ValueError
            If vectorizers are different between the two ``FeatureSet`` instances.

        ValueError
            If there are duplicate feature names.

        ValueError
            If there are conflicting labels.
        """
        # Check that the sets of IDs are equal
        if set(self.ids) != set(other.ids):
            raise ValueError("IDs are not in the same order in each " "feature set")
        # Compute the relative ordering of IDs for merging the features
        # and labels.
        ids_indices = dict((y, x) for x, y in enumerate(other.ids))
        relative_order = [ids_indices[self_id] for self_id in self.ids]

        # Initialize the new feature set with a name and the IDs.
        new_set = FeatureSet("+".join(sorted([self.name, other.name])), deepcopy(self.ids))

        # Make sure that features and vectorizer in either feature set are not None
        if (
            self.vectorizer is None
            or other.vectorizer is None
            or self.features is None
            or other.features is None
        ):
            raise ValueError(
                "Cannot combine FeatureSets since either the vectorizer "
                "or the features are not defined."
            )

        # Make sure the two vectorizers are the same type
        if not isinstance(self.vectorizer, type(other.vectorizer)):
            raise ValueError(
                "Cannot combine FeatureSets because they are "
                "not both using the same type of feature "
                "vectorizer (e.g., DictVectorizer, "
                "FeatureHasher)"
            )
        # they have to be the same types in this block
        else:
            uses_feature_hasher = isinstance(self.vectorizer, FeatureHasher)
            if uses_feature_hasher:
                if self.vectorizer.n_features != other.vectorizer.n_features:
                    raise ValueError(
                        "Cannot combine FeatureSets that use "
                        "FeatureHashers with different values of "
                        "n_features setting."
                    )
            else:
                # Check for duplicate feature names.
                if set(self.vectorizer.feature_names_) & set(other.vectorizer.feature_names_):
                    raise ValueError(
                        "Cannot combine FeatureSets because they have duplicate feature names."
                    )
            num_feats = self.features.shape[1]

            new_set.features = sp.hstack([self.features, other.features[relative_order]], "csr")
            new_set.vectorizer = deepcopy(self.vectorizer)
            if not uses_feature_hasher:
                for feat_name, index in other.vectorizer.vocabulary_.items():
                    new_set.vectorizer.vocabulary_[feat_name] = index + num_feats
                other_names = other.vectorizer.feature_names_
                new_set.vectorizer.feature_names_.extend(other_names)

            # If either set has labels, check that they don't conflict.
            if self.has_labels:
                # labels should be the same for each FeatureSet, so store once.
                conflicts = not np.all(self.labels == other.labels[relative_order])  # type: ignore
                if other.has_labels and conflicts:
                    raise ValueError(
                        "Feature sets have conflicting labels for examples with the same ID."
                    )
                new_set.labels = deepcopy(self.labels)
            else:
                labels = other.labels
                if other.has_labels:
                    labels = deepcopy(other.labels[relative_order])  # type: ignore
                new_set.labels = labels

        return new_set

    def filter(
        self,
        ids: Optional[List[IdType]] = None,
        labels: Optional[List[LabelType]] = None,
        features: Optional[List[str]] = None,
        inverse: bool = False,
    ) -> None:
        """
        Remove or keep features and/or examples from the given feature set.

        Filtering is done in-place.

        Parameters
        ----------
        ids : Optional[List[:class:`skll.types.IdType`]], default=None
            Examples to keep in the FeatureSet. If ``None``, no ID
            filtering takes place.

        labels : Optional[List[:class:`skll.types.LabelType`]], default=None
            Labels that we want to retain examples for. If ``None``,
            no label filtering takes place.

        features : Optional[List[str]], default=None
            Features to keep in the FeatureSet. To help with
            filtering string-valued features that were converted
            to sequences of boolean features when read in, any
            features in the FeatureSet that contain a ``=`` will be
            split on the first occurrence and the prefix will be
            checked to see if it is in ``features``.
            If ``None``, no feature filtering takes place.
            Cannot be used if FeatureSet uses a FeatureHasher for
            vectorization.

        inverse : bool, default=False
            Instead of keeping features and/or examples in lists,
            remove them.

        Raises
        ------
        ValueError
            If attempting to use features to filter a ``FeatureSet`` that
            uses a ``FeatureHasher`` vectorizer.
        """
        # Construct mask that indicates which examples to keep
        mask = np.ones(len(self), dtype=bool)
        if ids:
            mask = np.logical_and(mask, np.in1d(self.ids, ids))
        if labels and self.labels is not None:
            mask = np.logical_and(mask, np.in1d(self.labels, labels))

        if inverse and (labels is not None or ids is not None):
            mask = np.logical_not(mask)

        # Remove examples not in mask
        self.ids = self.ids[mask]
        if self.labels is not None:
            self.labels = self.labels[mask]
        if self.features is not None:
            self.features = self.features[mask, :]

        # Filter features
        if features and self.features is not None and self.vectorizer is not None:
            if isinstance(self.vectorizer, FeatureHasher):
                raise ValueError(
                    "FeatureSets with FeatureHasher vectorizers cannot be filtered by feature."
                )
            columns = np.array(
                sorted(
                    {
                        feat_num
                        for feat_name, feat_num in self.vectorizer.vocabulary_.items()
                        if (feat_name in features or feat_name.split("=", 1)[0] in features)
                    }
                )
            )
            if inverse:
                all_columns = np.arange(self.features.shape[1])
                columns = all_columns[np.logical_not(np.in1d(all_columns, columns))]
            self.features = self.features[:, columns]
            self.vectorizer.restrict(columns, indices=True)

    def filtered_iter(
        self,
        ids: Optional[List[IdType]] = None,
        labels: Optional[List[LabelType]] = None,
        features: Optional[Collection[str]] = None,
        inverse: bool = False,
    ) -> FeatGenerator:
        """
        Retain only the specified features and/or examples from the output.

        Parameters
        ----------
        ids : Optional[List[:class:`skll.types.IdType`]], default=None
            Examples to keep in the ``FeatureSet``. If ``None``, no ID
            filtering takes place.

        labels : Optional[List[:class:`skll.types.LabelType`]], default=None
            Labels that we want to retain examples for. If ``None``,
            no label filtering takes place.

        features : Optional[Collection[str]], default=None
            Features to keep in the ``FeatureSet``. To help with
            filtering string-valued features that were converted
            to sequences of boolean features when read in, any
            features in the ``FeatureSet`` that contain a `=` will be
            split on the first occurrence and the prefix will be
            checked to see if it is in ``features``.
            If `None`, no feature filtering takes place.
            Cannot be used if ``FeatureSet`` uses a FeatureHasher for
            vectorization.

        inverse : bool, default=False
            Instead of keeping features and/or examples in lists,
            remove them.

        Returns
        -------
        :class:`skll.types.FeatGenerator`

            A generator that yields 3-tuples containing:

              - :class:`skll.types.IdType`  - The ID of the example.

              - :class:`skll.types.LabelType` - The label of the example.

              - :class:`skll.types.FeatureDict` - The feature dictionary, with
                feature name as the key and example value as the value.

        Raises
        ------
        ValueError
            If the vectorizer is not a ``DictVectorizer``.

        ValueError
            If any of the "labels", "features", or "vectorizer" attribute
            is ``None``.
        """
        if self.features is not None and not isinstance(self.vectorizer, DictVectorizer):
            raise ValueError(
                "FeatureSets can only be iterated through if they"
                " use a DictVectorizer for their feature "
                "vectorizer."
            )

        if self.labels is None or self.features is None or self.vectorizer is None:
            raise ValueError("Cannot filter featureset with no labels, features, or vectorizer.")
        else:
            for id_, label_, feats in zip(self.ids, self.labels, self.features):
                # Skip instances with IDs not in filter
                if ids is not None and (id_ in ids) == inverse:
                    continue
                # Skip instances with labels not in filter
                if labels is not None and (label_ in labels) == inverse:
                    continue

                # reshape to a 2D matrix if we are not using a sparse matrix
                # to store the features
                feats = feats.reshape(1, -1) if not sp.issparse(feats) else feats
                feat_dict = self.vectorizer.inverse_transform(feats)[0]
                if features is not None:
                    feat_dict = {
                        name: value
                        for name, value in feat_dict.items()
                        if (inverse != (name in features or name.split("=", 1)[0] in features))
                    }
                elif not inverse:
                    feat_dict = {}
                yield id_, label_, feat_dict

    def __sub__(self, other: "FeatureSet") -> "FeatureSet":
        """
        Subset ``FeatureSet`` instance by removing all features from ``other`` instance.

        Parameters
        ----------
        other : :class:`skll.data.featureset.FeatureSet`
            The other ``FeatureSet`` containing the features that should
            be removed from this ``FeatureSet``.

        Returns
        -------
        :class:`skll.data.featureset.FeatureSet`
            A copy of ``self`` with all features in ``other`` removed.
        """
        new_set = deepcopy(self)
        if other.vectorizer:
            new_set.filter(features=other.vectorizer.feature_names_, inverse=True)
        return new_set

    @property
    def has_labels(self):
        """
        Check if ``FeatureSet`` has finite labels.

        Returns
        -------
        has_labels : bool
            Whether or not this FeatureSet has any finite labels.
        """
        # make sure that labels is not None or a list of Nones
        if self.labels is not None and not all(label is None for label in self.labels):
            # then check that they are not a list of NaNs
            return not (
                np.issubdtype(self.labels.dtype, np.floating) and np.isnan(np.min(self.labels))
            )
        else:
            return False

    def __str__(self):
        """
        Return a string representation of ``FeatureSet``.

        Returns
        -------
        str:
            A string representation of ``FeatureSet``.
        """
        return str(self.__dict__)

    def __repr__(self):
        """
        Return a string representation of ``FeatureSet``.

        Returns
        -------
        str:
            A string representation of ``FeatureSet``.
        """
        return repr(self.__dict__)

    def __getitem__(
        self, value: Union[int, slice]
    ) -> Union["FeatureSet", Tuple[IdType, LabelType, FeatureDictList]]:
        """
        Get new feature subset or specific example.

        Parameters
        ----------
        value: Union[int, slice]
            The value to use for retrieval. This can either be a slice or
            an index.

        Returns
        -------
        Union[:class:`skll.data.featureset.FeatureSet`, Tuple[:class:`skll.types.IdType`, :class:`skll.types.LabelType`, :class:`skll.types.FeatureDictList`]]  # noqa: E501
            If `value` is a slice, then return a new ``FeatureSet`` instance
            containing a subset of the data. If it's an index, return the
            specific example by row number.
        """
        # Check if we're slicing
        if isinstance(value, slice):
            sliced_ids = self.ids[value]
            sliced_feats = self.features[value] if self.features is not None else None
            sliced_labels = self.labels[value] if self.labels is not None else None
            return FeatureSet(
                f"{self.name}_{value}",
                sliced_ids,
                features=sliced_feats,
                labels=sliced_labels,
                vectorizer=self.vectorizer,
            )
        else:
            label = self.labels[value] if self.labels is not None else ""
            if self.features is not None and self.vectorizer:
                submatrix = self.features[value, :]
                features = self.vectorizer.inverse_transform(submatrix)[0]
            else:
                features = [{}]
            return self.ids[value], label, features

    @staticmethod
    def split(
        fs: "FeatureSet", ids_for_split1: List[int], ids_for_split2: Optional[List[int]] = None
    ) -> Tuple["FeatureSet", "FeatureSet"]:
        """
        Split ``FeatureSet`` into two new ``FeatureSet`` instances.

        The splitting is done based on the given indices for the two splits.

        Parameters
        ----------
        fs : skll.data.featureset.FeatureSet
            The ``FeatureSet`` instance to split.

        ids_for_split1 : List[int]
            A list of example indices which will be split out into
            the first ``FeatureSet`` instance. Note that the
            FeatureSet instance will respect the order of the
            specified indices.

        ids_for_split2 : Optional[List[int]], default=None
            An optional list of example indices which will be
            split out into the second ``FeatureSet`` instance.
            Note that the ``FeatureSet`` instance will respect
            the order of the specified indices. If this is
            not specified, then the second ``FeatureSet``
            instance will contain the complement of the
            first set of indices sorted in ascending order.

        Returns
        -------
        Tuple[:class:`skll.data.featureset.FeatureSet`, :class:`skll.data.featureset.FeatureSet`]
            A tuple containing the two featureset instances.
        """
        # Note: an alternative way to implement this is to make copies
        # of the given FeatureSet instance and then use the `filter()`
        # method but that wastes too much memory since it requires making
        # two copies of the original FeatureSet which may be huge. With
        # the current implementation, we are creating new objects but
        # they should be much smaller than the original FeatureSet.
        ids1 = fs.ids[ids_for_split1]
        labels1 = fs.labels[ids_for_split1] if fs.labels is not None else None
        features1 = fs.features[ids_for_split1] if fs.features is not None else None

        # if ids_for_split2 is not given, it will be the complement of ids_split1
        if ids_for_split2 is None:
            ids_for_split2 = [ind for ind in range(len(fs.ids)) if ind not in ids_for_split1]

        ids2 = fs.ids[ids_for_split2]
        labels2 = fs.labels[ids_for_split2] if fs.labels is not None else None
        features2 = fs.features[ids_for_split2] if fs.features is not None else None

        fs1 = FeatureSet(
            f"{fs.name}_1", ids1, labels=labels1, features=features1, vectorizer=fs.vectorizer
        )
        fs2 = FeatureSet(
            f"{fs.name}_2", ids2, labels=labels2, features=features2, vectorizer=fs.vectorizer
        )
        return fs1, fs2

    @staticmethod
    def from_data_frame(
        df: DataFrame,
        name: str,
        labels_column: Optional[str] = None,
        vectorizer: Optional[Union[DictVectorizer, FeatureHasher]] = None,
    ) -> "FeatureSet":
        """
        Create a ``FeatureSet`` instance from a pandas data frame.

        Will raise an Exception if pandas is not installed in your environment.
        The ``ids`` in the ``FeatureSet`` will be the index from the given frame.

        Parameters
        ----------
        df : pandas.DataFrame
            The pandas.DataFrame object to use as a ``FeatureSet``.

        name : str
            The name of the output ``FeatureSet`` instance.

        labels_column : Optional[str], default=None
            The name of the column containing the labels (data to predict).

        vectorizer : Optional[Union[:class:`sklearn.feature_extraction.DictVectorizer`, :class:`sklearn.feature_extraction.FeatureHasher`]], default=None
            Vectorizer which will be used to generate the feature matrix.

        Returns
        -------
        :class:`skll.data.featureset.FeatureSet`
            A ``FeatureSet`` instance generated from from the given data frame.
        """
        if labels_column:
            feature_columns = [column for column in df.columns if column != labels_column]
            labels = df[labels_column].tolist()
        else:
            feature_columns = df.columns
            labels = None

        features = df[feature_columns].to_dict(orient="records")
        return FeatureSet(
            name, ids=df.index.tolist(), labels=labels, features=features, vectorizer=vectorizer
        )



#!/usr/bin/env python
# License: BSD 3 clause
from setuptools import find_packages, setup

# Get version without importing, which avoids dependency issues
exec(compile(open("skll/version.py").read(), "skll/version.py", "exec"))
# (we use the above instead of execfile for Python 3.x compatibility)


def readme():
    with open("README.rst") as f:
        return f.read()


def requirements():
    req_path = "requirements.txt"
    with open(req_path) as f:
        reqs = f.read().splitlines()
    return reqs


setup(
    name="skll",
    version=__version__,  # type: ignore # noqa: F821
    description=(
        "SciKit-Learn Laboratory makes it easier to run machine "
        "learning experiments with scikit-learn."
    ),
    long_description=readme(),
    keywords="learning scikit-learn",
    url="http://github.com/EducationalTestingService/skll",
    author="Nitin Madnani",
    author_email="nmadnani@ets.org",
    license="BSD 3 clause",
    packages=find_packages(exclude=["tests", "examples"]),
    entry_points={
        "console_scripts": [
            "filter_features = skll.utils.commandline.filter_features:main",
            "generate_predictions = skll.utils.commandline.generate_predictions:main",
            "join_features = skll.utils.commandline.join_features:main",
            "print_model_weights = skll.utils.commandline.print_model_weights:main",
            "run_experiment = skll.utils.commandline.run_experiment:main",
            "skll_convert = skll.utils.commandline.skll_convert:main",
            "summarize_results = skll.utils.commandline.summarize_results:main",
            "compute_eval_from_predictions = skll.utils.commandline.compute_eval_from_predictions:main",  # noqa: E501
            "plot_learning_curves = skll.utils.commandline.plot_learning_curves:main",
        ]
    },
    install_requires=requirements(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    zip_safe=False,
)
